# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Any

import libcst as cst
from libcst import PartialParserConfig, parse_expression, parse_statement
from libcst._nodes.tests.base import CSTNodeTest
from libcst.metadata import CodeRange
from libcst.testing.utils import data_provider


class AwaitTest(CSTNodeTest):
    @data_provider(
        (
            # Some simple calls
            {
                "node": cst.Await(cst.Name("test")),
                "code": "await test",
                "parser": lambda code: parse_expression(
                    code, config=PartialParserConfig(python_version="3.7")
                ),
                "expected_position": None,
            },
            {
                "node": cst.Await(cst.Call(cst.Name("test"))),
                "code": "await test()",
                "parser": lambda code: parse_expression(
                    code, config=PartialParserConfig(python_version="3.7")
                ),
                "expected_position": None,
            },
            # Whitespace
            {
                "node": cst.Await(
                    cst.Name("test"),
                    whitespace_after_await=cst.SimpleWhitespace("  "),
                    lpar=(cst.LeftParen(whitespace_after=cst.SimpleWhitespace(" ")),),
                    rpar=(cst.RightParen(whitespace_before=cst.SimpleWhitespace(" ")),),
                ),
                "code": "( await  test )",
                "parser": lambda code: parse_expression(
                    code, config=PartialParserConfig(python_version="3.7")
                ),
                "expected_position": CodeRange((1, 2), (1, 13)),
            },
        )
    )
    def test_valid_py37(self, **kwargs: Any) -> None:
        # We don't have sentinel nodes for atoms, so we know that 100% of atoms
        # can be parsed identically to their creation.
        self.validate_node(**kwargs)

    @data_provider(
        (
            # Some simple calls
            {
                "node": cst.FunctionDef(
                    cst.Name("foo"),
                    cst.Parameters(),
                    cst.IndentedBlock(
                        (
                            cst.SimpleStatementLine(
                                (cst.Expr(cst.Await(cst.Name("test"))),)
                            ),
                        )
                    ),
                    asynchronous=cst.Asynchronous(),
                ),
                "code": "async def foo():\n    await test\n",
                "parser": lambda code: parse_statement(
                    code, config=PartialParserConfig(python_version="3.6")
                ),
                "expected_position": None,
            },
            {
                "node": cst.FunctionDef(
                    cst.Name("foo"),
                    cst.Parameters(),
                    cst.IndentedBlock(
                        (
                            cst.SimpleStatementLine(
                                (cst.Expr(cst.Await(cst.Call(cst.Name("test")))),)
                            ),
                        )
                    ),
                    asynchronous=cst.Asynchronous(),
                ),
                "code": "async def foo():\n    await test()\n",
                "parser": lambda code: parse_statement(
                    code, config=PartialParserConfig(python_version="3.6")
                ),
                "expected_position": None,
            },
            # Whitespace
            {
                "node": cst.FunctionDef(
                    cst.Name("foo"),
                    cst.Parameters(),
                    cst.IndentedBlock(
                        (
                            cst.SimpleStatementLine(
                                (
                                    cst.Expr(
                                        cst.Await(
                                            cst.Name("test"),
                                            whitespace_after_await=cst.SimpleWhitespace(
                                                "  "
                                            ),
                                            lpar=(
                                                cst.LeftParen(
                                                    whitespace_after=cst.SimpleWhitespace(
                                                        " "
                                                    )
                                                ),
                                            ),
                                            rpar=(
                                                cst.RightParen(
                                                    whitespace_before=cst.SimpleWhitespace(
                                                        " "
                                                    )
                                                ),
                                            ),
                                        )
                                    ),
                                )
                            ),
                        )
                    ),
                    asynchronous=cst.Asynchronous(),
                ),
                "code": "async def foo():\n    ( await  test )\n",
                "parser": lambda code: parse_statement(
                    code, config=PartialParserConfig(python_version="3.6")
                ),
                "expected_position": None,
            },
        )
    )
    def test_valid_py36(self, **kwargs: Any) -> None:
        # We don't have sentinel nodes for atoms, so we know that 100% of atoms
        # can be parsed identically to their creation.
        self.validate_node(**kwargs)

    @data_provider(
        (
            # Expression wrapping parenthesis rules
            {
                "get_node": (
                    lambda: cst.Await(cst.Name("foo"), lpar=(cst.LeftParen(),))
                ),
                "expected_re": "left paren without right paren",
            },
            {
                "get_node": (
                    lambda: cst.Await(cst.Name("foo"), rpar=(cst.RightParen(),))
                ),
                "expected_re": "right paren without left paren",
            },
            {
                "get_node": (
                    lambda: cst.Await(
                        cst.Name("foo"), whitespace_after_await=cst.SimpleWhitespace("")
                    )
                ),
                "expected_re": "at least one space after await",
            },
        )
    )
    def test_invalid(self, **kwargs: Any) -> None:
        self.assert_invalid(**kwargs)



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.


from textwrap import dedent
from typing import Callable

import libcst as cst
from libcst.testing.utils import UnitTest, data_provider


class ParseErrorsTest(UnitTest):
    @data_provider(
        {
            # _wrapped_tokenize raises these exceptions
            "wrapped_tokenize__invalid_token": (
                lambda: cst.parse_module("'"),
                dedent(
                    """
                    Syntax Error @ 1:1.
                    "'" is not a valid token.

                    '
                    ^
                    """
                ).strip(),
            ),
            "wrapped_tokenize__expected_dedent": (
                lambda: cst.parse_module("if False:\n    pass\n  pass"),
                dedent(
                    """
                    Syntax Error @ 3:1.
                    Inconsistent indentation. Expected a dedent.

                      pass
                    ^
                    """
                ).strip(),
            ),
            "wrapped_tokenize__mismatched_braces": (
                lambda: cst.parse_module("abcd)"),
                dedent(
                    """
                    Syntax Error @ 1:5.
                    Encountered a closing brace without a matching opening brace.

                    abcd)
                        ^
                    """
                ).strip(),
            ),
            # _base_parser raises these exceptions
            "base_parser__unexpected_indent": (
                lambda: cst.parse_module("    abcd"),
                dedent(
                    """
                    Syntax Error @ 1:5.
                    Incomplete input. Unexpectedly encountered an indent.

                        abcd
                        ^
                    """
                ).strip(),
            ),
            "base_parser__unexpected_dedent": (
                lambda: cst.parse_module("if False:\n    (el for el\n"),
                dedent(
                    """
                    Syntax Error @ 3:1.
                    Incomplete input. Encountered a dedent, but expected 'in'.

                        (el for el
                                  ^
                    """
                ).strip(),
            ),
            "base_parser__multiple_possibilities": (
                lambda: cst.parse_module("try: pass"),
                dedent(
                    """
                    Syntax Error @ 2:1.
                    Incomplete input. Encountered end of file (EOF), but expected 'except', or 'finally'.

                    try: pass
                             ^
                    """
                ).strip(),
            ),
            # conversion functions raise these exceptions.
            # `_base_parser` is responsible for attaching location information.
            "convert_nonterminal__dict_unpacking": (
                lambda: cst.parse_expression("{**el for el in []}"),
                dedent(
                    """
                    Syntax Error @ 1:19.
                    dict unpacking cannot be used in dict comprehension

                    {**el for el in []}
                                      ^
                    """
                ).strip(),
            ),
            "convert_nonterminal__arglist_non_default_after_default": (
                lambda: cst.parse_statement("def fn(first=None, second): ..."),
                dedent(
                    """
                    Syntax Error @ 1:26.
                    Cannot have a non-default argument following a default argument.

                    def fn(first=None, second): ...
                                             ^
                    """
                ).strip(),
            ),
            "convert_nonterminal__arglist_trailing_param_star_without_comma": (
                lambda: cst.parse_statement("def fn(abc, *): ..."),
                dedent(
                    """
                    Syntax Error @ 1:14.
                    Named (keyword) arguments must follow a bare *.

                    def fn(abc, *): ...
                                 ^
                    """
                ).strip(),
            ),
            "convert_nonterminal__arglist_trailing_param_star_with_comma": (
                lambda: cst.parse_statement("def fn(abc, *,): ..."),
                dedent(
                    """
                    Syntax Error @ 1:15.
                    Named (keyword) arguments must follow a bare *.

                    def fn(abc, *,): ...
                                  ^
                    """
                ).strip(),
            ),
            "convert_nonterminal__class_arg_positional_after_keyword": (
                lambda: cst.parse_statement("class Cls(first=None, second): ..."),
                dedent(
                    """
                    Syntax Error @ 2:1.
                    Positional argument follows keyword argument.

                    class Cls(first=None, second): ...
                                                      ^
                    """
                ).strip(),
            ),
            "convert_nonterminal__class_arg_positional_expansion_after_keyword": (
                lambda: cst.parse_statement("class Cls(first=None, *second): ..."),
                dedent(
                    """
                    Syntax Error @ 2:1.
                    Positional argument follows keyword argument.

                    class Cls(first=None, *second): ...
                                                       ^
                    """
                ).strip(),
            ),
        }
    )
    def test_parser_syntax_error_str(
        self, parse_fn: Callable[[], object], expected: str
    ) -> None:
        with self.assertRaises(cst.ParserSyntaxError) as cm:
            parse_fn()
        self.assertEqual(str(cm.exception), expected)



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import dataclasses

import libcst as cst
import libcst.matchers as m
from libcst.matchers import matches
from libcst.testing.utils import UnitTest


class MatchersMatcherTest(UnitTest):
    def test_simple_matcher_true(self) -> None:
        # Match based on identical attributes.
        self.assertTrue(matches(cst.Name("foo"), m.Name("foo")))

    def test_simple_matcher_false(self) -> None:
        # Fail to match due to incorrect value on Name.
        self.assertFalse(matches(cst.Name("foo"), m.Name("bar")))

    def test_complex_matcher_true(self) -> None:
        # Match on any Call, not caring about arguments.
        self.assertTrue(
            matches(
                cst.Call(
                    func=cst.Name("foo"),
                    args=(
                        cst.Arg(cst.Integer("1")),
                        cst.Arg(cst.Integer("2")),
                        cst.Arg(cst.Integer("3")),
                    ),
                ),
                m.Call(),
            )
        )
        # Match on any Call to a function named "foo".
        self.assertTrue(
            matches(
                cst.Call(
                    func=cst.Name("foo"),
                    args=(
                        cst.Arg(cst.Integer("1")),
                        cst.Arg(cst.Integer("2")),
                        cst.Arg(cst.Integer("3")),
                    ),
                ),
                m.Call(m.Name("foo")),
            )
        )
        # Match on any Call to a function named "foo" with three arguments.
        self.assertTrue(
            matches(
                cst.Call(
                    func=cst.Name("foo"),
                    args=(
                        cst.Arg(cst.Integer("1")),
                        cst.Arg(cst.Integer("2")),
                        cst.Arg(cst.Integer("3")),
                    ),
                ),
                m.Call(func=m.Name("foo"), args=(m.Arg(), m.Arg(), m.Arg())),
            )
        )
        # Match any Call to a function named "foo" with three integer arguments.
        self.assertTrue(
            matches(
                cst.Call(
                    func=cst.Name("foo"),
                    args=(
                        cst.Arg(cst.Integer("1")),
                        cst.Arg(cst.Integer("2")),
                        cst.Arg(cst.Integer("3")),
                    ),
                ),
                m.Call(
                    func=m.Name("foo"),
                    args=(m.Arg(m.Integer()), m.Arg(m.Integer()), m.Arg(m.Integer())),
                ),
            )
        )
        # Match any Call to a function named "foo" with integer arguments 1, 2, 3.
        self.assertTrue(
            matches(
                cst.Call(
                    func=cst.Name("foo"),
                    args=(
                        cst.Arg(cst.Integer("1")),
                        cst.Arg(cst.Integer("2")),
                        cst.Arg(cst.Integer("3")),
                    ),
                ),
                m.Call(
                    func=m.Name("foo"),
                    args=(
                        m.Arg(m.Integer("1")),
                        m.Arg(m.Integer("2")),
                        m.Arg(m.Integer("3")),
                    ),
                ),
            )
        )
        # Match any Call to a function named "foo" with three arguments, the last one
        # being the integer 3.
        self.assertTrue(
            matches(
                cst.Call(
                    func=cst.Name("foo"),
                    args=(
                        cst.Arg(cst.Integer("1")),
                        cst.Arg(cst.Integer("2")),
                        cst.Arg(cst.Integer("3")),
                    ),
                ),
                m.Call(
                    func=m.Name("foo"),
                    args=(m.DoNotCare(), m.DoNotCare(), m.Arg(m.Integer("3"))),
                ),
            )
        )

    def test_complex_matcher_false(self) -> None:
        # Fail to match since this is a Call, not a FunctionDef.
        self.assertFalse(
            matches(
                cst.Call(
                    func=cst.Name("foo"),
                    args=(
                        cst.Arg(cst.Integer("1")),
                        cst.Arg(cst.Integer("2")),
                        cst.Arg(cst.Integer("3")),
                    ),
                ),
                m.FunctionDef(),
            )
        )
        # Fail to match a function named "bar".
        self.assertFalse(
            matches(
                cst.Call(
                    func=cst.Name("foo"),
                    args=(
                        cst.Arg(cst.Integer("1")),
                        cst.Arg(cst.Integer("2")),
                        cst.Arg(cst.Integer("3")),
                    ),
                ),
                m.Call(m.Name("bar")),
            )
        )
        # Fail to match a function named "foo" with two arguments.
        self.assertFalse(
            matches(
                cst.Call(
                    func=cst.Name("foo"),
                    args=(
                        cst.Arg(cst.Integer("1")),
                        cst.Arg(cst.Integer("2")),
                        cst.Arg(cst.Integer("3")),
                    ),
                ),
                m.Call(func=m.Name("foo"), args=(m.Arg(), m.Arg())),
            )
        )
        # Fail to match a function named "foo" with three integer arguments
        # 3, 2, 1.
        self.assertFalse(
            matches(
                cst.Call(
                    func=cst.Name("foo"),
                    args=(
                        cst.Arg(cst.Integer("1")),
                        cst.Arg(cst.Integer("2")),
                        cst.Arg(cst.Integer("3")),
                    ),
                ),
                m.Call(
                    func=m.Name("foo"),
                    args=(
                        m.Arg(m.Integer("3")),
                        m.Arg(m.Integer("2")),
                        m.Arg(m.Integer("1")),
                    ),
                ),
            )
        )
        # Fail to match a function named "foo" with three arguments, the last one
        # being the integer 1.
        self.assertFalse(
            matches(
                cst.Call(
                    func=cst.Name("foo"),
                    args=(
                        cst.Arg(cst.Integer("1")),
                        cst.Arg(cst.Integer("2")),
                        cst.Arg(cst.Integer("3")),
                    ),
                ),
                m.Call(
                    func=m.Name("foo"),
                    args=(m.DoNotCare(), m.DoNotCare(), m.Arg(m.Integer("1"))),
                ),
            )
        )

    def test_type_of_matcher_true(self) -> None:
        self.assertTrue(matches(cst.Name("true"), m.TypeOf(m.Name)))
        self.assertTrue(matches(cst.Name("true"), m.TypeOf(m.Name)(value="true")))
        self.assertTrue(matches(cst.Name("true"), m.Name | m.Float | m.SimpleString))
        self.assertTrue(
            matches(cst.SimpleString("'foo'"), m.TypeOf(m.Name, m.SimpleString))
        )
        self.assertTrue(
            matches(
                cst.SimpleString("'foo'"),
                m.TypeOf(m.Name, m.SimpleString)(value="'foo'"),
            )
        )
        with self.assertRaises(Exception):
            # pyre-ignore
            m.TypeOf(cst.Float)(value=1.0) | cst.Name

        with self.assertRaises(TypeError):
            # pyre-ignore
            m.TypeOf(cst.Float) & cst.SimpleString

        for case in (
            cst.BinaryOperation(
                left=cst.Name("foo"), operator=cst.Add(), right=cst.Name("bar")
            ),
            cst.BooleanOperation(
                left=cst.Name("foo"), operator=cst.Or(), right=cst.Name("bar")
            ),
        ):
            self.assertTrue(
                matches(
                    case, (m.BinaryOperation | m.BooleanOperation)(left=m.Name("foo"))
                )
            )
            new_case = dataclasses.replace(case, left=case.right, right=case.left)
            self.assertTrue(
                matches(
                    new_case,
                    ~(m.BinaryOperation | m.BooleanOperation)(left=m.Name("foo")),
                )
            )

    def test_type_of_matcher_false(self) -> None:
        self.assertFalse(matches(cst.Name("true"), m.TypeOf(m.SimpleString)))
        self.assertFalse(matches(cst.Name("true"), m.TypeOf(m.Name)(value="false")))
        self.assertFalse(
            matches(cst.Name("true"), m.TypeOf(m.SimpleString)(value="true"))
        )
        self.assertFalse(
            matches(cst.SimpleString("'foo'"), m.TypeOf(m.Name, m.Attribute))
        )
        self.assertFalse(
            matches(
                cst.SimpleString("'foo'"), m.TypeOf(m.Name, m.Attribute)(value="'foo'")
            )
        )
        self.assertFalse(
            matches(
                cst.SimpleString("'foo'"),
                m.TypeOf(m.Name, m.SimpleString)(value="'bar'"),
            )
        )

        for case in (
            cst.BinaryOperation(
                left=cst.Name("foo"), operator=cst.Add(), right=cst.Name("bar")
            ),
            cst.BooleanOperation(
                left=cst.Name("foo"), operator=cst.Or(), right=cst.Name("bar")
            ),
        ):
            self.assertFalse(
                matches(
                    case, (m.BinaryOperation | m.BooleanOperation)(left=m.Name("bar"))
                )
            )
            self.assertFalse(
                matches(
                    case, ~(m.BinaryOperation | m.BooleanOperation)(left=m.Name("foo"))
                )
            )

    def test_or_matcher_true(self) -> None:
        # Match on either True or False identifier.
        self.assertTrue(
            matches(cst.Name("True"), m.OneOf(m.Name("True"), m.Name("False")))
        )
        # Match any assignment that assigns a value of True or False to an
        # unspecified target.
        self.assertTrue(
            matches(
                cst.Assign((cst.AssignTarget(cst.Name("x")),), cst.Name("True")),
                m.Assign(value=m.OneOf(m.Name("True"), m.Name("False"))),
            )
        )
        self.assertTrue(
            matches(
                cst.Call(
                    func=cst.Name("foo"),
                    args=(
                        cst.Arg(cst.Integer("1")),
                        cst.Arg(cst.Integer("2")),
                        cst.Arg(cst.Integer("3")),
                    ),
                ),
                m.Call(
                    func=m.Name("foo"),
                    args=m.OneOf(
                        (
                            m.Arg(m.Integer("3")),
                            m.Arg(m.Integer("2")),
                            m.Arg(m.Integer("1")),
                        ),
                        (
                            m.Arg(m.Integer("1")),
                            m.Arg(m.Integer("2")),
                            m.Arg(m.Integer("3")),
                        ),
                    ),
                ),
            )
        )

    def test_or_matcher_false(self) -> None:
        # Fail to match since None is not True or False.
        self.assertFalse(
            matches(cst.Name("None"), m.OneOf(m.Name("True"), m.Name("False")))
        )
        # Fail to match since assigning None to a target is not the same as
        # assigning True or False to a target.
        self.assertFalse(
            matches(
                cst.Assign((cst.AssignTarget(cst.Name("x")),), cst.Name("None")),
                m.Assign(value=m.OneOf(m.Name("True"), m.Name("False"))),
            )
        )
        self.assertFalse(
            matches(
                cst.Call(
                    func=cst.Name("foo"),
                    args=(
                        cst.Arg(cst.Integer("1")),
                        cst.Arg(cst.Integer("2")),
                        cst.Arg(cst.Integer("3")),
                    ),
                ),
                m.Call(
                    func=m.Name("foo"),
                    args=m.OneOf(
                        (
                            m.Arg(m.Integer("3")),
                            m.Arg(m.Integer("2")),
                            m.Arg(m.Integer("1")),
                        ),
                        (
                            m.Arg(m.Integer("4")),
                            m.Arg(m.Integer("5")),
                            m.Arg(m.Integer("6")),
                        ),
                    ),
                ),
            )
        )

    def test_or_operator_matcher_true(self) -> None:
        # Match on either True or False identifier.
        self.assertTrue(matches(cst.Name("True"), m.Name("True") | m.Name("False")))
        # Match on either True or False identifier.
        self.assertTrue(matches(cst.Name("False"), m.Name("True") | m.Name("False")))
        # Match on either True, False or None identifier.
        self.assertTrue(
            matches(cst.Name("None"), m.Name("True") | m.Name("False") | m.Name("None"))
        )
        # Match any assignment that assigns a value of True or False to an
        # unspecified target.
        self.assertTrue(
            matches(
                cst.Assign((cst.AssignTarget(cst.Name("x")),), cst.Name("True")),
                m.Assign(value=m.Name("True") | m.Name("False")),
            )
        )

    def test_or_operator_matcher_false(self) -> None:
        # Fail to match since None is not True or False.
        self.assertFalse(matches(cst.Name("None"), m.Name("True") | m.Name("False")))
        # Fail to match since assigning None to a target is not the same as
        # assigning True or False to a target.
        self.assertFalse(
            matches(
                cst.Assign((cst.AssignTarget(cst.Name("x")),), cst.Name("None")),
                m.Assign(value=m.Name("True") | m.Name("False")),
            )
        )

    def test_zero_or_more_matcher_no_args_true(self) -> None:
        # Match a function call to "foo" with any number of arguments as
        # long as the first one is an integer with the value 1.
        self.assertTrue(
            matches(
                cst.Call(
                    func=cst.Name("foo"),
                    args=(
                        cst.Arg(cst.Integer("1")),
                        cst.Arg(cst.Integer("2")),
                        cst.Arg(cst.Integer("3")),
                    ),
                ),
                m.Call(
                    func=m.Name("foo"), args=(m.Arg(m.Integer("1")), m.ZeroOrMore())
                ),
            )
        )
        # Match a function call to "foo" with any number of arguments as
        # long as one of them is an integer with the value 1.
        self.assertTrue(
            matches(
                cst.Call(
                    func=cst.Name("foo"),
                    args=(
                        cst.Arg(cst.Integer("1")),
                        cst.Arg(cst.Integer("2")),
                        cst.Arg(cst.Integer("3")),
                    ),
                ),
                m.Call(
                    func=m.Name("foo"),
                    args=(m.ZeroOrMore(), m.Arg(m.Integer("1")), m.ZeroOrMore()),
                ),
            )
        )
        # Match a function call to "foo" with any number of arguments as
        # long as one of them is an integer with the value 2.
        self.assertTrue(
            matches(
                cst.Call(
                    func=cst.Name("foo"),
                    args=(
                        cst.Arg(cst.Integer("1")),
                        cst.Arg(cst.Integer("2")),
                        cst.Arg(cst.Integer("3")),
                    ),
                ),
                m.Call(
                    func=m.Name("foo"),
                    args=(m.ZeroOrMore(), m.Arg(m.Integer("2")), m.ZeroOrMore()),
                ),
            )
        )
        # Match a function call to "foo" with any number of arguments as
        # long as one of them is an integer with the value 3.
        self.assertTrue(
            matches(
                cst.Call(
                    func=cst.Name("foo"),
                    args=(
                        cst.Arg(cst.Integer("1")),
                        cst.Arg(cst.Integer("2")),
                        cst.Arg(cst.Integer("3")),
                    ),
                ),
                m.Call(
                    func=m.Name("foo"),
                    args=(m.ZeroOrMore(), m.Arg(m.Integer("3")), m.ZeroOrMore()),
                ),
            )
        )
        # Match a function call to "foo" with any number of arguments as
        # long as the last one is an integer with the value 3.
        self.assertTrue(
            matches(
                cst.Call(
                    func=cst.Name("foo"),
                    args=(
                        cst.Arg(cst.Integer("1")),
                        cst.Arg(cst.Integer("2")),
                        cst.Arg(cst.Integer("3")),
                    ),
                ),
                m.Call(
                    func=m.Name("foo"), args=(m.ZeroOrMore(), m.Arg(m.Integer("3")))
                ),
            )
        )
        # Match a function call to "foo" with any number of arguments as
        # long as there are two arguments with the values 1 and 3 anywhere
        # in the argument list, respecting order.
        self.assertTrue(
            matches(
                cst.Call(
                    func=cst.Name("foo"),
                    args=(
                        cst.Arg(cst.Integer("1")),
                        cst.Arg(cst.Integer("2")),
                        cst.Arg(cst.Integer("3")),
                    ),
                ),
                m.Call(
                    func=m.Name("foo"),
                    args=(
                        m.ZeroOrMore(),
                        m.Arg(m.Integer("1")),
                        m.ZeroOrMore(),
                        m.Arg(m.Integer("3")),
                        m.ZeroOrMore(),
                    ),
                ),
            )
        )
        # Match a function call to "foo" with any number of arguments as
        # long as there are three arguments with the values 1, 2 and 3 anywhere
        # in the argument list, respecting order.
        self.assertTrue(
            matches(
                cst.Call(
                    func=cst.Name("foo"),
                    args=(
                        cst.Arg(cst.Integer("1")),
                        cst.Arg(cst.Integer("2")),
                        cst.Arg(cst.Integer("3")),
                    ),
                ),
                m.Call(
                    func=m.Name("foo"),
                    args=(
                        m.ZeroOrMore(),
                        m.Arg(m.Integer("1")),
                        m.ZeroOrMore(),
                        m.Arg(m.Integer("2")),
                        m.ZeroOrMore(),
                        m.Arg(m.Integer("3")),
                        m.ZeroOrMore(),
                    ),
                ),
            )
        )

    def test_at_least_n_matcher_no_args_true(self) -> None:
        # Match a function call to "foo" with at least one argument.
        self.assertTrue(
            matches(
                cst.Call(
                    func=cst.Name("foo"),
                    args=(
                        cst.Arg(cst.Integer("1")),
                        cst.Arg(cst.Integer("2")),
                        cst.Arg(cst.Integer("3")),
                    ),
                ),
                m.Call(func=m.Name("foo"), args=(m.AtLeastN(n=1),)),
            )
        )
        # Match a function call to "foo" with at least two arguments.
        self.assertTrue(
            matches(
                cst.Call(
                    func=cst.Name("foo"),
                    args=(
                        cst.Arg(cst.Integer("1")),
                        cst.Arg(cst.Integer("2")),
                        cst.Arg(cst.Integer("3")),
                    ),
                ),
                m.Call(func=m.Name("foo"), args=(m.AtLeastN(n=2),)),
            )
        )
        # Match a function call to "foo" with at least three arguments.
        self.assertTrue(
            matches(
                cst.Call(
                    func=cst.Name("foo"),
                    args=(
                        cst.Arg(cst.Integer("1")),
                        cst.Arg(cst.Integer("2")),
                        cst.Arg(cst.Integer("3")),
                    ),
                ),
                m.Call(func=m.Name("foo"), args=(m.AtLeastN(n=3),)),
            )
        )
        # Match a function call to "foo" with at least two arguments the
        # first one being the integer 1.
        self.assertTrue(
            matches(
                cst.Call(
                    func=cst.Name("foo"),
                    args=(
                        cst.Arg(cst.Integer("1")),
                        cst.Arg(cst.Integer("2")),
                        cst.Arg(cst.Integer("3")),
                    ),
                ),
                m.Call(
                    func=m.Name("foo"), args=(m.Arg(m.Integer("1")), m.AtLeastN(n=1))
                ),
            )
        )
        # Match a function call to "foo" with at least three arguments the
        # first one being the integer 1.
        self.assertTrue(
            matches(
                cst.Call(
                    func=cst.Name("foo"),
                    args=(
                        cst.Arg(cst.Integer("1")),
                        cst.Arg(cst.Integer("2")),
                        cst.Arg(cst.Integer("3")),
                    ),
                ),
                m.Call(
                    func=m.Name("foo"), args=(m.Arg(m.Integer("1")), m.AtLeastN(n=2))
                ),
            )
        )
        # Match a function call to "foo" with at least three arguments. The
        # There should be an argument with the value 2, which should have
        # at least one argument before and one argument after.
        self.assertTrue(
            matches(
                cst.Call(
                    func=cst.Name("foo"),
                    args=(
                        cst.Arg(cst.Integer("1")),
                        cst.Arg(cst.Integer("2")),
                        cst.Arg(cst.Integer("3")),
                    ),
                ),
                m.Call(
                    func=m.Name("foo"),
                    args=(m.AtLeastN(n=1), m.Arg(m.Integer("2")), m.AtLeastN(n=1)),
                ),
            )
        )
        # Match a function call to "foo" with at least two arguments, the last
        # one being the value 3.
        self.assertTrue(
            matches(
                cst.Call(
                    func=cst.Name("foo"),
                    args=(
                        cst.Arg(cst.Integer("1")),
                        cst.Arg(cst.Integer("2")),
                        cst.Arg(cst.Integer("3")),
                    ),
                ),
                m.Call(
                    func=m.Name("foo"), args=(m.AtLeastN(n=1), m.Arg(m.Integer("3")))
                ),
            )
        )
        # Match a function call to "foo" with at least three arguments, the last
        # one being the value 3.
        self.assertTrue(
            matches(
                cst.Call(
                    func=cst.Name("foo"),
                    args=(
                        cst.Arg(cst.Integer("1")),
                        cst.Arg(cst.Integer("2")),
                        cst.Arg(cst.Integer("3")),
                    ),
                ),
                m.Call(
                    func=m.Name("foo"), args=(m.AtLeastN(n=2), m.Arg(m.Integer("3")))
                ),
            )
        )

    def test_at_least_n_matcher_no_args_false(self) -> None:
        # Fail to match a function call to "foo" with at least four arguments.
        self.assertFalse(
            matches(
                cst.Call(
                    func=cst.Name("foo"),
                    args=(
                        cst.Arg(cst.Integer("1")),
                        cst.Arg(cst.Integer("2")),
                        cst.Arg(cst.Integer("3")),
                    ),
                ),
                m.Call(func=m.Name("foo"), args=(m.AtLeastN(n=4),)),
            )
        )
        # Fail to match a function call to "foo" with at least four arguments,
        # the first one being the value 1.
        self.assertFalse(
            matches(
                cst.Call(
                    func=cst.Name("foo"),
                    args=(
                        cst.Arg(cst.Integer("1")),
                        cst.Arg(cst.Integer("2")),
                        cst.Arg(cst.Integer("3")),
                    ),
                ),
                m.Call(
                    func=m.Name("foo"), args=(m.Arg(m.Integer("1")), m.AtLeastN(n=3))
                ),
            )
        )
        # Fail to match a function call to "foo" with at least three arguments,
        # the last one being the value 2.
        self.assertFalse(
            matches(
                cst.Call(
                    func=cst.Name("foo"),
                    args=(
                        cst.Arg(cst.Integer("1")),
                        cst.Arg(cst.Integer("2")),
                        cst.Arg(cst.Integer("3")),
                    ),
                ),
                m.Call(
                    func=m.Name("foo"), args=(m.AtLeastN(n=2), m.Arg(m.Integer("2")))
                ),
            )
        )

    def test_zero_or_more_matcher_args_true(self) -> None:
        # Match a function call to "foo" where the first argument is the integer
        # value 1, and the rest of the arguements are wildcards.
        self.assertTrue(
            matches(
                cst.Call(
                    func=cst.Name("foo"),
                    args=(
                        cst.Arg(cst.Integer("1")),
                        cst.Arg(cst.Integer("2")),
                        cst.Arg(cst.Integer("3")),
                    ),
                ),
                m.Call(
                    func=m.Name("foo"),
                    args=(m.Arg(m.Integer("1")), m.ZeroOrMore(m.Arg())),
                ),
            )
        )
        # Match a function call to "foo" where the first argument is the integer
        # value 1, and the rest of the arguements are integers of any value.
        self.assertTrue(
            matches(
                cst.Call(
                    func=cst.Name("foo"),
                    args=(
                        cst.Arg(cst.Integer("1")),
                        cst.Arg(cst.Integer("2")),
                        cst.Arg(cst.Integer("3")),
                    ),
                ),
                m.Call(
                    func=m.Name("foo"),
                    args=(m.Arg(m.Integer("1")), m.ZeroOrMore(m.Arg(m.Integer()))),
                ),
            )
        )
        # Match a function call to "foo" with zero or more arguments, where the
        # first argument can optionally be the integer 1 or 2, and the second
        # can only be the integer 2. This case verifies non-greedy behavior in the
        # matcher.
        self.assertTrue(
            matches(
                cst.Call(
                    func=cst.Name("foo"),
                    args=(
                        cst.Arg(cst.Integer("1")),
                        cst.Arg(cst.Integer("2")),
                        cst.Arg(cst.Integer("3")),
                    ),
                ),
                m.Call(
                    func=m.Name("foo"),
                    args=(
                        m.ZeroOrMore(m.Arg(m.OneOf(m.Integer("1"), m.Integer("2")))),
                        m.Arg(m.Integer("2")),
                        m.ZeroOrMore(),
                    ),
                ),
            )
        )
        # Match a function call to "foo" where the first argument is the integer
        # value 1, and the rest of the arguements are integers with the value
        # 2 or 3.
        self.assertTrue(
            matches(
                cst.Call(
                    func=cst.Name("foo"),
                    args=(
                        cst.Arg(cst.Integer("1")),
                        cst.Arg(cst.Integer("2")),
                        cst.Arg(cst.Integer("3")),
                    ),
                ),
                m.Call(
                    func=m.Name("foo"),
                    args=(
                        m.Arg(m.Integer("1")),
                        m.ZeroOrMore(m.Arg(m.OneOf(m.Integer("2"), m.Integer("3")))),
                    ),
                ),
            )
        )

    def test_zero_or_more_matcher_args_false(self) -> None:
        # Fail to match a function call to "foo" where the first argument is the
        # integer value 1, and the rest of the arguments are strings.
        self.assertFalse(
            matches(
                cst.Call(
                    func=cst.Name("foo"),
                    args=(
                        cst.Arg(cst.Integer("1")),
                        cst.Arg(cst.Integer("2")),
                        cst.Arg(cst.Integer("3")),
                    ),
                ),
                m.Call(
                    func=m.Name("foo"),
                    args=(m.Arg(m.Integer("1")), m.ZeroOrMore(m.Arg(m.SimpleString()))),
                ),
            )
        )
        # Fail to match a function call to "foo" where the first argument is the
        # integer value 1, and the rest of the arguements are integers with the
        # value 2.
        self.assertFalse(
            matches(
                cst.Call(
                    func=cst.Name("foo"),
                    args=(
                        cst.Arg(cst.Integer("1")),
                        cst.Arg(cst.Integer("2")),
                        cst.Arg(cst.Integer("3")),
                    ),
                ),
                m.Call(
                    func=m.Name("foo"),
                    args=(m.Arg(m.Integer("1")), m.ZeroOrMore(m.Arg(m.Integer("2")))),
                ),
            )
        )

    def test_at_least_n_matcher_args_true(self) -> None:
        # Match a function call to "foo" where the first argument is the integer
        # value 1, and there are at least two wildcard arguments after.
        self.assertTrue(
            matches(
                cst.Call(
                    func=cst.Name("foo"),
                    args=(
                        cst.Arg(cst.Integer("1")),
                        cst.Arg(cst.Integer("2")),
                        cst.Arg(cst.Integer("3")),
                    ),
                ),
                m.Call(
                    func=m.Name("foo"),
                    args=(m.Arg(m.Integer("1")), m.AtLeastN(m.Arg(), n=2)),
                ),
            )
        )
        # Match a function call to "foo" where the first argument is the integer
        # value 1, and there are at least two arguements are integers of any value
        # after.
        self.assertTrue(
            matches(
                cst.Call(
                    func=cst.Name("foo"),
                    args=(
                        cst.Arg(cst.Integer("1")),
                        cst.Arg(cst.Integer("2")),
                        cst.Arg(cst.Integer("3")),
                    ),
                ),
                m.Call(
                    func=m.Name("foo"),
                    args=(m.Arg(m.Integer("1")), m.AtLeastN(m.Arg(m.Integer()), n=2)),
                ),
            )
        )
        # Match a function call to "foo" where the first argument is the integer
        # value 1, and there are at least two arguements that are integers with the
        # value 2 or 3 after.
        self.assertTrue(
            matches(
                cst.Call(
                    func=cst.Name("foo"),
                    args=(
                        cst.Arg(cst.Integer("1")),
                        cst.Arg(cst.Integer("2")),
                        cst.Arg(cst.Integer("3")),
                    ),
                ),
                m.Call(
                    func=m.Name("foo"),
                    args=(
                        m.Arg(m.Integer("1")),
                        m.AtLeastN(m.Arg(m.OneOf(m.Integer("2"), m.Integer("3"))), n=2),
                    ),
                ),
            )
        )

    def test_at_least_n_matcher_args_false(self) -> None:
        # Fail to match a function call to "foo" where the first argument is the
        # integer value 1, and there are at least two arguments after that are
        # strings.
        self.assertFalse(
            matches(
                cst.Call(
                    func=cst.Name("foo"),
                    args=(
                        cst.Arg(cst.Integer("1")),
                        cst.Arg(cst.Integer("2")),
                        cst.Arg(cst.Integer("3")),
                    ),
                ),
                m.Call(
                    func=m.Name("foo"),
                    args=(
                        m.Arg(m.Integer("1")),
                        m.AtLeastN(m.Arg(m.SimpleString()), n=2),
                    ),
                ),
            )
        )
        # Fail to match a function call to "foo" where the first argument is the integer
        # value 1, and there are at least three wildcard arguments after.
        self.assertFalse(
            matches(
                cst.Call(
                    func=cst.Name("foo"),
                    args=(
                        cst.Arg(cst.Integer("1")),
                        cst.Arg(cst.Integer("2")),
                        cst.Arg(cst.Integer("3")),
                    ),
                ),
                m.Call(
                    func=m.Name("foo"),
                    args=(m.Arg(m.Integer("1")), m.AtLeastN(m.Arg(), n=3)),
                ),
            )
        )
        # Fail to match a function call to "foo" where the first argument is the
        # integer value 1, and there are at least two arguements that are integers with
        # the value 2 after.
        self.assertFalse(
            matches(
                cst.Call(
                    func=cst.Name("foo"),
                    args=(
                        cst.Arg(cst.Integer("1")),
                        cst.Arg(cst.Integer("2")),
                        cst.Arg(cst.Integer("3")),
                    ),
                ),
                m.Call(
                    func=m.Name("foo"),
                    args=(
                        m.Arg(m.Integer("1")),
                        m.AtLeastN(m.Arg(m.Integer("2")), n=2),
                    ),
                ),
            )
        )

    def test_at_most_n_matcher_no_args_true(self) -> None:
        # Match a function call to "foo" with at most two arguments.
        self.assertTrue(
            matches(
                cst.Call(func=cst.Name("foo"), args=(cst.Arg(cst.Integer("1")),)),
                m.Call(func=m.Name("foo"), args=(m.AtMostN(n=2),)),
            )
        )
        # Match a function call to "foo" with at most two arguments.
        self.assertTrue(
            matches(
                cst.Call(
                    func=cst.Name("foo"),
                    args=(cst.Arg(cst.Integer("1")), cst.Arg(cst.Integer("2"))),
                ),
                m.Call(func=m.Name("foo"), args=(m.AtMostN(n=2),)),
            )
        )
        # Match a function call to "foo" with at most six arguments, the last
        # one being the integer 1.
        self.assertTrue(
            matches(
                cst.Call(func=cst.Name("foo"), args=(cst.Arg(cst.Integer("1")),)),
                m.Call(
                    func=m.Name("foo"), args=[m.AtMostN(n=5), m.Arg(m.Integer("1"))]
                ),
            )
        )
        # Match a function call to "foo" with at most six arguments, the last
        # one being the integer 1.
        self.assertTrue(
            matches(
                cst.Call(
                    func=cst.Name("foo"),
                    args=(cst.Arg(cst.Integer("1")), cst.Arg(cst.Integer("2"))),
                ),
                m.Call(
                    func=m.Name("foo"), args=(m.AtMostN(n=5), m.Arg(m.Integer("2")))
                ),
            )
        )
        # Match a function call to "foo" with at most six arguments, the first
        # one being the integer 1.
        self.assertTrue(
            matches(
                cst.Call(
                    func=cst.Name("foo"),
                    args=(cst.Arg(cst.Integer("1")), cst.Arg(cst.Integer("2"))),
                ),
                m.Call(
                    func=m.Name("foo"), args=(m.Arg(m.Integer("1")), m.AtMostN(n=5))
                ),
            )
        )
        # Match a function call to "foo" with at most six arguments, the first
        # one being the integer 1.
        self.assertTrue(
            matches(
                cst.Call(
                    func=cst.Name("foo"),
                    args=(cst.Arg(cst.Integer("1")), cst.Arg(cst.Integer("2"))),
                ),
                m.Call(func=m.Name("foo"), args=(m.Arg(m.Integer("1")), m.ZeroOrOne())),
            )
        )

    def test_at_most_n_matcher_no_args_false(self) -> None:
        # Fail to match a function call to "foo" with at most two arguments.
        self.assertFalse(
            matches(
                cst.Call(
                    func=cst.Name("foo"),
                    args=(
                        cst.Arg(cst.Integer("1")),
                        cst.Arg(cst.Integer("2")),
                        cst.Arg(cst.Integer("3")),
                    ),
                ),
                m.Call(func=m.Name("foo"), args=(m.AtMostN(n=2),)),
            )
        )
        # Fail to match a function call to "foo" with at most two arguments,
        # the last one being the integer 3.
        self.assertFalse(
            matches(
                cst.Call(
                    func=cst.Name("foo"),
                    args=(
                        cst.Arg(cst.Integer("1")),
                        cst.Arg(cst.Integer("2")),
                        cst.Arg(cst.Integer("3")),
                    ),
                ),
                m.Call(
                    func=m.Name("foo"), args=(m.AtMostN(n=1), m.Arg(m.Integer("3")))
                ),
            )
        )
        # Fail to match a function call to "foo" with at most two arguments,
        # the last one being the integer 3.
        self.assertFalse(
            matches(
                cst.Call(
                    func=cst.Name("foo"),
                    args=(
                        cst.Arg(cst.Integer("1")),
                        cst.Arg(cst.Integer("2")),
                        cst.Arg(cst.Integer("3")),
                    ),
                ),
                m.Call(func=m.Name("foo"), args=(m.ZeroOrOne(), m.Arg(m.Integer("3")))),
            )
        )

    def test_at_most_n_matcher_args_true(self) -> None:
        # Match a function call to "foo" with at most two arguments, both of which
        # are the integer 1.
        self.assertTrue(
            matches(
                cst.Call(func=cst.Name("foo"), args=(cst.Arg(cst.Integer("1")),)),
                m.Call(
                    func=m.Name("foo"), args=(m.AtMostN(m.Arg(m.Integer("1")), n=2),)
                ),
            )
        )
        # Match a function call to "foo" with at most two arguments, both of which
        # can be the integer 1 or 2.
        self.assertTrue(
            matches(
                cst.Call(
                    func=cst.Name("foo"),
                    args=(cst.Arg(cst.Integer("1")), cst.Arg(cst.Integer("2"))),
                ),
                m.Call(
                    func=m.Name("foo"),
                    args=(
                        m.AtMostN(m.Arg(m.OneOf(m.Integer("1"), m.Integer("2"))), n=2),
                    ),
                ),
            )
        )
        # Match a function call to "foo" with at most two arguments, the first
        # one being the integer 1 and the second one, if included, being the
        # integer 2.
        self.assertTrue(
            matches(
                cst.Call(
                    func=cst.Name("foo"),
                    args=(cst.Arg(cst.Integer("1")), cst.Arg(cst.Integer("2"))),
                ),
                m.Call(
                    func=m.Name("foo"),
                    args=(m.Arg(m.Integer("1")), m.ZeroOrOne(m.Arg(m.Integer("2")))),
                ),
            )
        )
        # Match a function call to "foo" with at most six arguments, the first
        # one being the integer 1 and the second one, if included, being the
        # integer 2.
        self.assertTrue(
            matches(
                cst.Call(
                    func=cst.Name("foo"),
                    args=(cst.Arg(cst.Integer("1")), cst.Arg(cst.Integer("2"))),
                ),
                m.Call(
                    func=m.Name("foo"),
                    args=(m.Arg(m.Integer("1")), m.ZeroOrOne(m.Arg(m.Integer("2")))),
                ),
            )
        )

    def test_at_most_n_matcher_args_false(self) -> None:
        # Fail to match a function call to "foo" with at most three arguments,
        # all of which are the integer 4.
        self.assertFalse(
            matches(
                cst.Call(
                    func=cst.Name("foo"),
                    args=(
                        cst.Arg(cst.Integer("1")),
                        cst.Arg(cst.Integer("2")),
                        cst.Arg(cst.Integer("3")),
                    ),
                ),
                m.Call(
                    func=m.Name("foo"), args=(m.AtMostN(m.Arg(m.Integer("4")), n=3),)
                ),
            )
        )

    def test_lambda_matcher_true(self) -> None:
        # Match based on identical attributes.
        self.assertTrue(
            matches(
                cst.Name("foo"), m.Name(value=m.MatchIfTrue(lambda value: "o" in value))
            )
        )

    def test_lambda_matcher_false(self) -> None:
        # Fail to match due to incorrect value on Name.
        self.assertFalse(
            matches(
                cst.Name("foo"), m.Name(value=m.MatchIfTrue(lambda value: "a" in value))
            )
        )

    def test_regex_matcher_true(self) -> None:
        # Match based on identical attributes.
        self.assertTrue(matches(cst.Name("foo"), m.Name(value=m.MatchRegex(r".*o.*"))))

    def test_regex_matcher_false(self) -> None:
        # Fail to match due to incorrect value on Name.
        self.assertFalse(matches(cst.Name("foo"), m.Name(value=m.MatchRegex(r".*a.*"))))

    def test_and_matcher_true(self) -> None:
        # Match on True identifier in roundabout way.
        self.assertTrue(
            matches(
                cst.Name("True"), m.AllOf(m.Name(), m.Name(value=m.MatchRegex(r"True")))
            )
        )
        self.assertTrue(
            matches(
                cst.Call(
                    func=cst.Name("foo"),
                    args=(
                        cst.Arg(cst.Integer("1")),
                        cst.Arg(cst.Integer("2")),
                        cst.Arg(cst.Integer("3")),
                    ),
                ),
                m.Call(
                    func=m.Name("foo"),
                    args=m.AllOf(
                        (m.Arg(), m.Arg(), m.Arg()),
                        (
                            m.Arg(m.Integer("1")),
                            m.Arg(m.Integer("2")),
                            m.Arg(m.Integer("3")),
                        ),
                    ),
                ),
            )
        )

    def test_and_matcher_false(self) -> None:
        # Fail to match since True and False cannot match.
        self.assertFalse(
            matches(cst.Name("None"), m.AllOf(m.Name("True"), m.Name("False")))
        )
        self.assertFalse(
            matches(
                cst.Call(
                    func=cst.Name("foo"),
                    args=(
                        cst.Arg(cst.Integer("1")),
                        cst.Arg(cst.Integer("2")),
                        cst.Arg(cst.Integer("3")),
                    ),
                ),
                m.Call(
                    func=m.Name("foo"),
                    args=m.AllOf(
                        (m.Arg(), m.Arg(), m.Arg()),
                        (
                            m.Arg(m.Integer("3")),
                            m.Arg(m.Integer("2")),
                            m.Arg(m.Integer("1")),
                        ),
                    ),
                ),
            )
        )

    def test_and_operator_matcher_true(self) -> None:
        # Match on True identifier in roundabout way.
        self.assertTrue(
            matches(cst.Name("True"), m.Name() & m.Name(value=m.MatchRegex(r"True")))
        )
        # Match in a really roundabout way that verifies the __or__ behavior on
        # AllOf itself.
        self.assertTrue(
            matches(
                cst.Name("True"),
                m.Name() & m.Name(value=m.MatchRegex(r"True")) & m.Name("True"),
            )
        )
        # Verify that MatchIfTrue works with __and__ behavior properly.
        self.assertTrue(
            matches(
                cst.Name("True"),
                m.MatchIfTrue(lambda x: isinstance(x, cst.Name))
                & m.Name(value=m.MatchRegex(r"True")),
            )
        )
        self.assertTrue(
            matches(
                cst.Name("True"),
                m.Name(value=m.MatchRegex(r"True"))
                & m.MatchIfTrue(lambda x: isinstance(x, cst.Name)),
            )
        )

    def test_and_operator_matcher_false(self) -> None:
        # Fail to match since True and False cannot match.
        self.assertFalse(matches(cst.Name("None"), m.Name("True") & m.Name("False")))

    def test_does_not_match_true(self) -> None:
        # Match on any call that takes one argument that isn't the value None.
        self.assertTrue(
            matches(
                cst.Call(func=cst.Name("foo"), args=(cst.Arg(cst.Name("True")),)),
                m.Call(args=(m.Arg(value=m.DoesNotMatch(m.Name("None"))),)),
            )
        )
        self.assertTrue(
            matches(
                cst.Call(func=cst.Name("foo"), args=(cst.Arg(cst.Integer("1")),)),
                m.Call(args=(m.DoesNotMatch(m.Arg(m.Name("None"))),)),
            )
        )
        self.assertTrue(
            matches(
                cst.Call(func=cst.Name("foo"), args=(cst.Arg(cst.Integer("1")),)),
                m.Call(args=m.DoesNotMatch((m.Arg(m.Integer("2")),))),
            )
        )
        # Match any call that takes an argument which isn't True or False.
        self.assertTrue(
            matches(
                cst.Call(func=cst.Name("foo"), args=(cst.Arg(cst.Integer("1")),)),
                m.Call(
                    args=(
                        m.Arg(
                            value=m.DoesNotMatch(
                                m.OneOf(m.Name("True"), m.Name("False"))
                            )
                        ),
                    )
                ),
            )
        )
        # Match any name node that doesn't match the regex for True
        self.assertTrue(
            matches(
                cst.Name("False"), m.Name(value=m.DoesNotMatch(m.MatchRegex(r"True")))
            )
        )

    def test_does_not_match_operator_true(self) -> None:
        # Match on any call that takes one argument that isn't the value None.
        self.assertTrue(
            matches(
                cst.Call(func=cst.Name("foo"), args=(cst.Arg(cst.Name("True")),)),
                m.Call(args=(m.Arg(value=~(m.Name("None"))),)),
            )
        )
        self.assertTrue(
            matches(
                cst.Call(func=cst.Name("foo"), args=(cst.Arg(cst.Integer("1")),)),
                m.Call(args=(~(m.Arg(m.Name("None"))),)),
            )
        )
        # Match any call that takes an argument which isn't True or False.
        self.assertTrue(
            matches(
                cst.Call(func=cst.Name("foo"), args=(cst.Arg(cst.Integer("1")),)),
                m.Call(args=(m.Arg(value=~(m.Name("True") | m.Name("False"))),)),
            )
        )
        self.assertTrue(
            matches(
                cst.Call(func=cst.Name("foo"), args=(cst.Arg(cst.Name("None")),)),
                m.Call(args=(m.Arg(value=(~(m.Name("True"))) & (~(m.Name("False")))),)),
            )
        )
        # Roundabout way to verify that or operator works with inverted nodes.
        self.assertTrue(
            matches(
                cst.Call(func=cst.Name("foo"), args=(cst.Arg(cst.Name("False")),)),
                m.Call(args=(m.Arg(value=(~(m.Name("True"))) | (~(m.Name("True")))),)),
            )
        )
        # Roundabout way to verify that inverse operator works properly on AllOf.
        self.assertTrue(
            matches(
                cst.Call(func=cst.Name("foo"), args=(cst.Arg(cst.Integer("1")),)),
                m.Call(args=(m.Arg(value=~(m.Name() & m.Name("True"))),)),
            )
        )
        # Match any name node that doesn't match the regex for True
        self.assertTrue(
            matches(cst.Name("False"), m.Name(value=~(m.MatchRegex(r"True"))))
        )

    def test_does_not_match_false(self) -> None:
        # Match on any call that takes one argument that isn't the value None.
        self.assertFalse(
            matches(
                cst.Call(func=cst.Name("foo"), args=(cst.Arg(cst.Name("None")),)),
                m.Call(args=(m.Arg(value=m.DoesNotMatch(m.Name("None"))),)),
            )
        )
        self.assertFalse(
            matches(
                cst.Call(func=cst.Name("foo"), args=(cst.Arg(cst.Integer("1")),)),
                m.Call(args=(m.DoesNotMatch(m.Arg(m.Integer("1"))),)),
            )
        )
        self.assertFalse(
            matches(
                cst.Call(func=cst.Name("foo"), args=(cst.Arg(cst.Integer("1")),)),
                m.Call(args=m.DoesNotMatch((m.Arg(m.Integer("1")),))),
            )
        )
        # Match any call that takes an argument which isn't True or False.
        self.assertFalse(
            matches(
                cst.Call(func=cst.Name("foo"), args=(cst.Arg(cst.Name("False")),)),
                m.Call(
                    args=(
                        m.Arg(
                            value=m.DoesNotMatch(
                                m.OneOf(m.Name("True"), m.Name("False"))
                            )
                        ),
                    )
                ),
            )
        )
        self.assertFalse(
            matches(
                cst.Call(func=cst.Name("foo"), args=(cst.Arg(cst.Name("True")),)),
                m.Call(args=(m.Arg(value=(~(m.Name("True"))) & (~(m.Name("False")))),)),
            )
        )
        # Match any name node that doesn't match the regex for True
        self.assertFalse(
            matches(
                cst.Name("True"), m.Name(value=m.DoesNotMatch(m.MatchRegex(r"True")))
            )
        )

    def test_does_not_match_operator_false(self) -> None:
        # Match on any call that takes one argument that isn't the value None.
        self.assertFalse(
            matches(
                cst.Call(func=cst.Name("foo"), args=(cst.Arg(cst.Name("None")),)),
                m.Call(args=(m.Arg(value=~(m.Name("None"))),)),
            )
        )
        self.assertFalse(
            matches(
                cst.Call(func=cst.Name("foo"), args=(cst.Arg(cst.Integer("1")),)),
                m.Call(args=((~(m.Arg(m.Integer("1")))),)),
            )
        )
        # Match any call that takes an argument which isn't True or False.
        self.assertFalse(
            matches(
                cst.Call(func=cst.Name("foo"), args=(cst.Arg(cst.Name("False")),)),
                m.Call(args=(m.Arg(value=~(m.Name("True") | m.Name("False"))),)),
            )
        )
        # Roundabout way of verifying ~(x&y) behavior.
        self.assertFalse(
            matches(
                cst.Call(func=cst.Name("foo"), args=(cst.Arg(cst.Name("False")),)),
                m.Call(args=(m.Arg(value=~(m.Name() & m.Name("False"))),)),
            )
        )
        # Roundabout way of verifying (~x)|(~y) behavior
        self.assertFalse(
            matches(
                cst.Call(func=cst.Name("foo"), args=(cst.Arg(cst.Name("True")),)),
                m.Call(args=(m.Arg(value=(~(m.Name("True"))) | (~(m.Name("True")))),)),
            )
        )
        # Match any name node that doesn't match the regex for True
        self.assertFalse(
            matches(cst.Name("True"), m.Name(value=~(m.MatchRegex(r"True"))))
        )

    def test_inverse_inverse_is_identity(self) -> None:
        # Verify that we don't wrap an InverseOf in an InverseOf in normal circumstances.
        identity = m.Name("True")
        self.assertTrue(m.DoesNotMatch(m.DoesNotMatch(identity)) is identity)
        self.assertTrue((~(~identity)) is identity)



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Any

import libcst as cst
from libcst import parse_statement
from libcst._nodes.tests.base import CSTNodeTest, DummyIndentedBlock
from libcst.metadata import CodeRange
from libcst.testing.utils import data_provider


class IfTest(CSTNodeTest):
    @data_provider(
        (
            # Simple if without elif or else
            {
                "node": cst.If(
                    cst.Name("conditional"), cst.SimpleStatementSuite((cst.Pass(),))
                ),
                "code": "if conditional: pass\n",
                "parser": parse_statement,
                "expected_position": CodeRange((1, 0), (1, 20)),
            },
            # else clause
            {
                "node": cst.If(
                    cst.Name("conditional"),
                    cst.SimpleStatementSuite((cst.Pass(),)),
                    orelse=cst.Else(cst.SimpleStatementSuite((cst.Pass(),))),
                ),
                "code": "if conditional: pass\nelse: pass\n",
                "parser": parse_statement,
                "expected_position": CodeRange((1, 0), (2, 10)),
            },
            # elif clause
            {
                "node": cst.If(
                    cst.Name("conditional"),
                    cst.SimpleStatementSuite((cst.Pass(),)),
                    orelse=cst.If(
                        cst.Name("other_conditional"),
                        cst.SimpleStatementSuite((cst.Pass(),)),
                        orelse=cst.Else(cst.SimpleStatementSuite((cst.Pass(),))),
                    ),
                ),
                "code": "if conditional: pass\nelif other_conditional: pass\nelse: pass\n",
                "parser": parse_statement,
                "expected_position": CodeRange((1, 0), (3, 10)),
            },
            # indentation
            {
                "node": DummyIndentedBlock(
                    "    ",
                    cst.If(
                        cst.Name("conditional"),
                        cst.SimpleStatementSuite((cst.Pass(),)),
                        orelse=cst.Else(cst.SimpleStatementSuite((cst.Pass(),))),
                    ),
                ),
                "code": "    if conditional: pass\n    else: pass\n",
                "parser": None,
                "expected_position": CodeRange((1, 4), (2, 14)),
            },
            # with an indented body
            {
                "node": DummyIndentedBlock(
                    "    ",
                    cst.If(
                        cst.Name("conditional"),
                        cst.IndentedBlock((cst.SimpleStatementLine((cst.Pass(),)),)),
                    ),
                ),
                "code": "    if conditional:\n        pass\n",
                "parser": None,
                "expected_position": CodeRange((1, 4), (2, 12)),
            },
            # leading_lines
            {
                "node": cst.If(
                    cst.Name("conditional"),
                    cst.SimpleStatementSuite((cst.Pass(),)),
                    leading_lines=(
                        cst.EmptyLine(comment=cst.Comment("# leading comment")),
                    ),
                ),
                "code": "# leading comment\nif conditional: pass\n",
                "parser": parse_statement,
                "expected_position": CodeRange((2, 0), (2, 20)),
            },
            # whitespace before/after test and else
            {
                "node": cst.If(
                    cst.Name("conditional"),
                    cst.SimpleStatementSuite((cst.Pass(),)),
                    whitespace_before_test=cst.SimpleWhitespace("   "),
                    whitespace_after_test=cst.SimpleWhitespace("  "),
                    orelse=cst.Else(
                        cst.SimpleStatementSuite((cst.Pass(),)),
                        whitespace_before_colon=cst.SimpleWhitespace(" "),
                    ),
                ),
                "code": "if   conditional  : pass\nelse : pass\n",
                "parser": parse_statement,
                "expected_position": CodeRange((1, 0), (2, 11)),
            },
            # empty lines between if/elif/else clauses, not captured by the suite.
            {
                "node": cst.If(
                    cst.Name("test_a"),
                    cst.SimpleStatementSuite((cst.Pass(),)),
                    orelse=cst.If(
                        cst.Name("test_b"),
                        cst.SimpleStatementSuite((cst.Pass(),)),
                        leading_lines=(cst.EmptyLine(),),
                        orelse=cst.Else(
                            cst.SimpleStatementSuite((cst.Pass(),)),
                            leading_lines=(cst.EmptyLine(),),
                        ),
                    ),
                ),
                "code": "if test_a: pass\n\nelif test_b: pass\n\nelse: pass\n",
                "parser": parse_statement,
                "expected_position": CodeRange((1, 0), (5, 10)),
            },
        )
    )
    def test_valid(self, **kwargs: Any) -> None:
        self.validate_node(**kwargs)



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Any

import libcst as cst
from libcst import parse_statement
from libcst._nodes.tests.base import CSTNodeTest
from libcst.helpers import ensure_type
from libcst.metadata import CodeRange
from libcst.testing.utils import data_provider


class RaiseConstructionTest(CSTNodeTest):
    @data_provider(
        (
            # Simple raise
            {"node": cst.Raise(), "code": "raise"},
            # Raise exception
            {
                "node": cst.Raise(cst.Call(cst.Name("Exception"))),
                "code": "raise Exception()",
                "expected_position": CodeRange((1, 0), (1, 17)),
            },
            # Raise exception from cause
            {
                "node": cst.Raise(
                    cst.Call(cst.Name("Exception")), cst.From(cst.Name("cause"))
                ),
                "code": "raise Exception() from cause",
            },
            # Whitespace oddities test
            {
                "node": cst.Raise(
                    cst.Call(
                        cst.Name("Exception"),
                        lpar=(cst.LeftParen(),),
                        rpar=(cst.RightParen(),),
                    ),
                    cst.From(
                        cst.Name(
                            "cause", lpar=(cst.LeftParen(),), rpar=(cst.RightParen(),)
                        ),
                        whitespace_before_from=cst.SimpleWhitespace(""),
                        whitespace_after_from=cst.SimpleWhitespace(""),
                    ),
                    whitespace_after_raise=cst.SimpleWhitespace(""),
                ),
                "code": "raise(Exception())from(cause)",
                "expected_position": CodeRange((1, 0), (1, 29)),
            },
            {
                "node": cst.Raise(
                    cst.Call(cst.Name("Exception")),
                    cst.From(
                        cst.Name("cause"),
                        whitespace_before_from=cst.SimpleWhitespace(""),
                    ),
                ),
                "code": "raise Exception()from cause",
                "expected_position": CodeRange((1, 0), (1, 27)),
            },
            # Whitespace rendering test
            {
                "node": cst.Raise(
                    exc=cst.Call(cst.Name("Exception")),
                    cause=cst.From(
                        cst.Name("cause"),
                        whitespace_before_from=cst.SimpleWhitespace("  "),
                        whitespace_after_from=cst.SimpleWhitespace("  "),
                    ),
                    whitespace_after_raise=cst.SimpleWhitespace("  "),
                ),
                "code": "raise  Exception()  from  cause",
                "expected_position": CodeRange((1, 0), (1, 31)),
            },
        )
    )
    def test_valid(self, **kwargs: Any) -> None:
        self.validate_node(**kwargs)

    @data_provider(
        (
            # Validate construction
            {
                "get_node": lambda: cst.Raise(cause=cst.From(cst.Name("cause"))),
                "expected_re": "Must have an 'exc' when specifying 'clause'. on Raise",
            },
            # Validate whitespace handling
            {
                "get_node": lambda: cst.Raise(
                    cst.Call(cst.Name("Exception")),
                    whitespace_after_raise=cst.SimpleWhitespace(""),
                ),
                "expected_re": "Must have at least one space after 'raise'",
            },
            {
                "get_node": lambda: cst.Raise(
                    cst.Name("exc"),
                    cst.From(
                        cst.Name("cause"),
                        whitespace_before_from=cst.SimpleWhitespace(""),
                    ),
                ),
                "expected_re": "Must have at least one space before 'from'",
            },
            {
                "get_node": lambda: cst.Raise(
                    cst.Name("exc"),
                    cst.From(
                        cst.Name("cause"),
                        whitespace_after_from=cst.SimpleWhitespace(""),
                    ),
                ),
                "expected_re": "Must have at least one space after 'from'",
            },
        )
    )
    def test_invalid(self, **kwargs: Any) -> None:
        self.assert_invalid(**kwargs)


class RaiseParsingTest(CSTNodeTest):
    @data_provider(
        (
            # Simple raise
            {"node": cst.Raise(), "code": "raise"},
            # Raise exception
            {
                "node": cst.Raise(
                    cst.Call(cst.Name("Exception")),
                    whitespace_after_raise=cst.SimpleWhitespace(" "),
                ),
                "code": "raise Exception()",
            },
            # Raise exception from cause
            {
                "node": cst.Raise(
                    cst.Call(cst.Name("Exception")),
                    cst.From(
                        cst.Name("cause"),
                        whitespace_before_from=cst.SimpleWhitespace(" "),
                        whitespace_after_from=cst.SimpleWhitespace(" "),
                    ),
                    whitespace_after_raise=cst.SimpleWhitespace(" "),
                ),
                "code": "raise Exception() from cause",
            },
            # Whitespace oddities test
            {
                "node": cst.Raise(
                    cst.Call(
                        cst.Name("Exception"),
                        lpar=(cst.LeftParen(),),
                        rpar=(cst.RightParen(),),
                    ),
                    cst.From(
                        cst.Name(
                            "cause", lpar=(cst.LeftParen(),), rpar=(cst.RightParen(),)
                        ),
                        whitespace_before_from=cst.SimpleWhitespace(""),
                        whitespace_after_from=cst.SimpleWhitespace(""),
                    ),
                    whitespace_after_raise=cst.SimpleWhitespace(""),
                ),
                "code": "raise(Exception())from(cause)",
            },
            {
                "node": cst.Raise(
                    cst.Call(cst.Name("Exception")),
                    cst.From(
                        cst.Name("cause"),
                        whitespace_before_from=cst.SimpleWhitespace(""),
                        whitespace_after_from=cst.SimpleWhitespace(" "),
                    ),
                    whitespace_after_raise=cst.SimpleWhitespace(" "),
                ),
                "code": "raise Exception()from cause",
            },
            # Whitespace rendering test
            {
                "node": cst.Raise(
                    exc=cst.Call(cst.Name("Exception")),
                    cause=cst.From(
                        cst.Name("cause"),
                        whitespace_before_from=cst.SimpleWhitespace("  "),
                        whitespace_after_from=cst.SimpleWhitespace("  "),
                    ),
                    whitespace_after_raise=cst.SimpleWhitespace("  "),
                ),
                "code": "raise  Exception()  from  cause",
            },
        )
    )
    def test_valid(self, **kwargs: Any) -> None:
        self.validate_node(
            parser=lambda code: ensure_type(
                parse_statement(code), cst.SimpleStatementLine
            ).body[0],
            **kwargs,
        )



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.


"""
Parso's tokenize doesn't give us tokens in the format that we'd ideally like, so this
performs a small number of transformations to the token stream:

- `end_pos` is precomputed as a property, instead of lazily as a method, for more
  efficient access.
- `whitespace_before` and `whitespace_after` have been added. These include the correct
  indentation information.
- `prefix` is removed, since we don't use it anywhere.
- `ERRORTOKEN` and `ERROR_DEDENT` have been removed, because we don't intend to support
  error recovery. If we encounter token errors, we'll raise a ParserSyntaxError instead.

If performance becomes a concern, we can rewrite this later as a fork of the original
tokenize module, instead of as a wrapper.
"""

from dataclasses import dataclass, field
from enum import Enum
from typing import Generator, List, Optional, Sequence

from libcst._add_slots import add_slots
from libcst._exceptions import ParserSyntaxError
from libcst._parser.parso.python.token import PythonTokenTypes, TokenType
from libcst._parser.parso.python.tokenize import (
    Token as OrigToken,
    tokenize_lines as orig_tokenize_lines,
)
from libcst._parser.parso.utils import PythonVersionInfo, split_lines
from libcst._parser.types.token import Token
from libcst._parser.types.whitespace_state import WhitespaceState


_ERRORTOKEN: TokenType = PythonTokenTypes.ERRORTOKEN
_ERROR_DEDENT: TokenType = PythonTokenTypes.ERROR_DEDENT

_INDENT: TokenType = PythonTokenTypes.INDENT
_DEDENT: TokenType = PythonTokenTypes.DEDENT
_ENDMARKER: TokenType = PythonTokenTypes.ENDMARKER

_FSTRING_START: TokenType = PythonTokenTypes.FSTRING_START
_FSTRING_END: TokenType = PythonTokenTypes.FSTRING_END

_OP: TokenType = PythonTokenTypes.OP


class _ParenthesisOrFStringStackEntry(Enum):
    PARENTHESIS = 0
    FSTRING = 0


_PARENTHESIS_STACK_ENTRY: _ParenthesisOrFStringStackEntry = (
    _ParenthesisOrFStringStackEntry.PARENTHESIS
)
_FSTRING_STACK_ENTRY: _ParenthesisOrFStringStackEntry = (
    _ParenthesisOrFStringStackEntry.FSTRING
)


@add_slots
@dataclass(frozen=False)
class _TokenizeState:
    lines: Sequence[str]
    previous_whitespace_state: WhitespaceState = field(
        default_factory=lambda: WhitespaceState(
            line=1, column=0, absolute_indent="", is_parenthesized=False
        )
    )
    indents: List[str] = field(default_factory=lambda: [""])
    parenthesis_or_fstring_stack: List[_ParenthesisOrFStringStackEntry] = field(
        default_factory=list
    )


def tokenize(
    code: str, version_info: PythonVersionInfo
) -> Generator[Token, None, None]:
    lines = split_lines(code, keepends=True)
    return tokenize_lines(lines, version_info)


def tokenize_lines(
    lines: Sequence[str], version_info: PythonVersionInfo
) -> Generator[Token, None, None]:
    state = _TokenizeState(lines)
    orig_tokens_iter = iter(orig_tokenize_lines(lines, version_info))

    # Iterate over the tokens and pass them to _convert_token, providing a one-token
    # lookahead, to enable proper indent handling.
    try:
        curr_token = next(orig_tokens_iter)
    except StopIteration:
        pass  # empty file
    else:
        for next_token in orig_tokens_iter:
            yield _convert_token(state, curr_token, next_token)
            curr_token = next_token
        yield _convert_token(state, curr_token, None)


def _convert_token(  # noqa: C901: too complex
    state: _TokenizeState, curr_token: OrigToken, next_token: Optional[OrigToken]
) -> Token:
    ct_type = curr_token.type
    ct_string = curr_token.string
    ct_start_pos = curr_token.start_pos
    if ct_type is _ERRORTOKEN:
        raise ParserSyntaxError(
            f"{ct_string!r} is not a valid token.",
            lines=state.lines,
            raw_line=ct_start_pos[0],
            raw_column=ct_start_pos[1],
        )
    if ct_type is _ERROR_DEDENT:
        raise ParserSyntaxError(
            "Inconsistent indentation. Expected a dedent.",
            lines=state.lines,
            raw_line=ct_start_pos[0],
            raw_column=ct_start_pos[1],
        )

    # Compute relative indent changes for indent/dedent nodes
    relative_indent: Optional[str] = None
    if ct_type is _INDENT:
        old_indent = "" if len(state.indents) < 2 else state.indents[-2]
        new_indent = state.indents[-1]
        relative_indent = new_indent[len(old_indent) :]

    if next_token is not None:
        nt_type = next_token.type
        if nt_type is _INDENT:
            nt_line, nt_column = next_token.start_pos
            state.indents.append(state.lines[nt_line - 1][:nt_column])
        elif nt_type is _DEDENT:
            state.indents.pop()

    whitespace_before = state.previous_whitespace_state

    if ct_type is _INDENT or ct_type is _DEDENT or ct_type is _ENDMARKER:
        # Don't update whitespace state for these dummy tokens. This makes it possible
        # to partially parse whitespace for IndentedBlock footers, and then parse the
        # rest of the whitespace in the following statement's leading_lines.
        # Unfortunately, that means that the indentation is either wrong for the footer
        # comments, or for the next line. We've chosen to allow it to be wrong for the
        # IndentedBlock footer and manually override the state when parsing whitespace
        # in that particular node.
        whitespace_after = whitespace_before
        ct_end_pos = ct_start_pos
    else:
        # Not a dummy token, so update the whitespace state.

        # Compute our own end_pos, since parso's end_pos is wrong for triple-strings.
        lines = split_lines(ct_string)
        if len(lines) > 1:
            ct_end_pos = ct_start_pos[0] + len(lines) - 1, len(lines[-1])
        else:
            ct_end_pos = (ct_start_pos[0], ct_start_pos[1] + len(ct_string))

        # Figure out what mode the whitespace parser should use. If we're inside
        # parentheses, certain whitespace (e.g. newlines) are allowed where they would
        # otherwise not be. f-strings override and disable this behavior, however.
        #
        # Parso's tokenizer tracks this internally, but doesn't expose it, so we have to
        # duplicate that logic here.

        pof_stack = state.parenthesis_or_fstring_stack
        try:
            if ct_type is _FSTRING_START:
                pof_stack.append(_FSTRING_STACK_ENTRY)
            elif ct_type is _FSTRING_END:
                pof_stack.pop()
            elif ct_type is _OP:
                if ct_string in "([{":
                    pof_stack.append(_PARENTHESIS_STACK_ENTRY)
                elif ct_string in ")]}":
                    pof_stack.pop()
        except IndexError:
            # pof_stack may be empty by the time we need to read from it due to
            # mismatched braces.
            raise ParserSyntaxError(
                "Encountered a closing brace without a matching opening brace.",
                lines=state.lines,
                raw_line=ct_start_pos[0],
                raw_column=ct_start_pos[1],
            )
        is_parenthesized = (
            len(pof_stack) > 0 and pof_stack[-1] == _PARENTHESIS_STACK_ENTRY
        )

        whitespace_after = WhitespaceState(
            ct_end_pos[0], ct_end_pos[1], state.indents[-1], is_parenthesized
        )

    # Hold onto whitespace_after, so we can use it as whitespace_before in the next
    # node.
    state.previous_whitespace_state = whitespace_after

    return Token(
        ct_type,
        ct_string,
        ct_start_pos,
        ct_end_pos,
        whitespace_before,
        whitespace_after,
        relative_indent,
    )



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import List

import libcst as cst
from libcst import CSTTransformer, CSTVisitor, parse_module
from libcst.testing.utils import UnitTest


class VisitorTest(UnitTest):
    def test_visitor(self) -> None:
        class SomeVisitor(CSTVisitor):
            def __init__(self) -> None:
                self.visit_order: List[str] = []

            def visit_If(self, node: cst.If) -> None:
                self.visit_order.append("visit_If")

            def leave_If(self, original_node: cst.If) -> None:
                self.visit_order.append("leave_If")

            def visit_If_test(self, node: cst.If) -> None:
                self.visit_order.append("visit_If_test")

            def leave_If_test(self, node: cst.If) -> None:
                self.visit_order.append("leave_If_test")

            def visit_Name(self, node: cst.Name) -> None:
                self.visit_order.append("visit_Name")

            def leave_Name(self, original_node: cst.Name) -> None:
                self.visit_order.append("leave_Name")

        # Create and visit a simple module.
        module = parse_module("if True:\n    pass")
        visitor = SomeVisitor()
        module.visit(visitor)

        # Check that visits worked.
        self.assertEqual(
            visitor.visit_order,
            [
                "visit_If",
                "visit_If_test",
                "visit_Name",
                "leave_Name",
                "leave_If_test",
                "leave_If",
            ],
        )

    def test_transformer(self) -> None:
        class SomeTransformer(CSTTransformer):
            def __init__(self) -> None:
                self.visit_order: List[str] = []

            def visit_If(self, node: cst.If) -> None:
                self.visit_order.append("visit_If")

            def leave_If(self, original_node: cst.If, updated_node: cst.If) -> cst.If:
                self.visit_order.append("leave_If")
                return updated_node

            def visit_If_test(self, node: cst.If) -> None:
                self.visit_order.append("visit_If_test")

            def leave_If_test(self, node: cst.If) -> None:
                self.visit_order.append("leave_If_test")

            def visit_Name(self, node: cst.Name) -> None:
                self.visit_order.append("visit_Name")

            def leave_Name(
                self, original_node: cst.Name, updated_node: cst.Name
            ) -> cst.Name:
                self.visit_order.append("leave_Name")
                return updated_node

        # Create and visit a simple module.
        module = parse_module("if True:\n    pass")
        transformer = SomeTransformer()
        module.visit(transformer)

        # Check that visits worked.
        self.assertEqual(
            transformer.visit_order,
            [
                "visit_If",
                "visit_If_test",
                "visit_Name",
                "leave_Name",
                "leave_If_test",
                "leave_If",
            ],
        )



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.


import abc
import builtins
from collections import defaultdict
from contextlib import contextmanager
from dataclasses import dataclass
from enum import Enum, auto
from typing import (
    Collection,
    Dict,
    Iterator,
    List,
    Mapping,
    MutableMapping,
    Optional,
    Set,
    Tuple,
    Type,
    Union,
)

import libcst as cst
from libcst import ensure_type
from libcst._add_slots import add_slots
from libcst._metadata_dependent import MetadataDependent
from libcst.helpers import get_full_name_for_node
from libcst.metadata.base_provider import BatchableMetadataProvider
from libcst.metadata.expression_context_provider import (
    ExpressionContext,
    ExpressionContextProvider,
)


# Comprehensions are handled separately in _visit_comp_alike due to
# the complexity of the semantics
_ASSIGNMENT_LIKE_NODES = (
    cst.AnnAssign,
    cst.AsName,
    cst.Assign,
    cst.AugAssign,
    cst.ClassDef,
    cst.CompFor,
    cst.FunctionDef,
    cst.Global,
    cst.Import,
    cst.ImportFrom,
    cst.NamedExpr,
    cst.Nonlocal,
    cst.Parameters,
    cst.WithItem,
)


@add_slots
@dataclass(frozen=False)
class Access:
    """
    An Access records an access of an assignment.

    .. note::
       This scope analysis only analyzes access via a :class:`~libcst.Name` or  a :class:`~libcst.Name`
       node embedded in other node like :class:`~libcst.Call` or :class:`~libcst.Attribute`.
       It doesn't support type annontation using :class:`~libcst.SimpleString` literal for forward
       references. E.g. in this example, the ``"Tree"`` isn't parsed as an access::

           class Tree:
               def __new__(cls) -> "Tree":
                   ...
    """

    #: The node of the access. A name is an access when the expression context is
    #: :attr:`ExpressionContext.LOAD`. This is usually the name node representing the
    #: access, except for dotted imports, when it might be the attribute that
    #: represents the most specific part of the imported symbol.
    node: Union[cst.Name, cst.Attribute]

    #: The scope of the access. Note that a access could be in a child scope of its
    #: assignment.
    scope: "Scope"

    is_annotation: bool

    is_type_hint: bool

    __assignments: Set["BaseAssignment"]
    __index: int

    def __init__(
        self, node: cst.Name, scope: "Scope", is_annotation: bool, is_type_hint: bool
    ) -> None:
        self.node = node
        self.scope = scope
        self.is_annotation = is_annotation
        self.is_type_hint = is_type_hint
        self.__assignments = set()
        self.__index = scope._assignment_count

    def __hash__(self) -> int:
        return id(self)

    @property
    def referents(self) -> Collection["BaseAssignment"]:
        """Return all assignments of the access."""
        return self.__assignments

    @property
    def _index(self) -> int:
        return self.__index

    def record_assignment(self, assignment: "BaseAssignment") -> None:
        if assignment.scope != self.scope or assignment._index < self.__index:
            self.__assignments.add(assignment)

    def record_assignments(self, name: str) -> None:
        assignments = self.scope[name]
        # filter out assignments that happened later than this access
        previous_assignments = {
            assignment
            for assignment in assignments
            if assignment.scope != self.scope or assignment._index < self.__index
        }
        if not previous_assignments and assignments and self.scope.parent != self.scope:
            previous_assignments = self.scope.parent[name]
        self.__assignments |= previous_assignments


class BaseAssignment(abc.ABC):
    """Abstract base class of :class:`Assignment` and :class:`BuitinAssignment`."""

    #: The name of assignment.
    name: str

    #: The scope associates to assignment.
    scope: "Scope"
    __accesses: Set[Access]

    def __init__(self, name: str, scope: "Scope") -> None:
        self.name = name
        self.scope = scope
        self.__accesses = set()

    def record_access(self, access: Access) -> None:
        if access.scope != self.scope or self._index < access._index:
            self.__accesses.add(access)

    def record_accesses(self, accesses: Set[Access]) -> None:
        later_accesses = {
            access
            for access in accesses
            if access.scope != self.scope or self._index < access._index
        }
        self.__accesses |= later_accesses
        earlier_accesses = accesses - later_accesses
        if earlier_accesses and self.scope.parent != self.scope:
            # Accesses "earlier" than the relevant assignment should be attached
            # to assignments of the same name in the parent
            for shadowed_assignment in self.scope.parent[self.name]:
                shadowed_assignment.record_accesses(earlier_accesses)

    @property
    def references(self) -> Collection[Access]:
        """Return all accesses of the assignment."""
        # we don't want to publicly expose the mutable version of this
        return self.__accesses

    def __hash__(self) -> int:
        return id(self)

    @property
    def _index(self) -> int:
        """Return an integer that represents the order of assignments in `scope`"""
        return -1


class Assignment(BaseAssignment):
    """An assignment records the name, CSTNode and its accesses."""

    #: The node of assignment, it could be a :class:`~libcst.Import`, :class:`~libcst.ImportFrom`,
    #: :class:`~libcst.Name`, :class:`~libcst.FunctionDef`, or :class:`~libcst.ClassDef`.
    node: cst.CSTNode
    __index: int

    def __init__(
        self, name: str, scope: "Scope", node: cst.CSTNode, index: int
    ) -> None:
        self.node = node
        self.__index = index
        super().__init__(name, scope)

    @property
    def _index(self) -> int:
        return self.__index


# even though we don't override the constructor.
class BuiltinAssignment(BaseAssignment):
    """
    A BuiltinAssignment represents an value provide by Python as a builtin, including
    `functions <https://docs.python.org/3/library/functions.html>`_,
    `constants <https://docs.python.org/3/library/constants.html>`_, and
    `types <https://docs.python.org/3/library/stdtypes.html>`_.
    """

    pass


class Assignments:
    """A container to provide all assignments in a scope."""

    def __init__(self, assignments: Mapping[str, Collection[BaseAssignment]]) -> None:
        self._assignments = assignments

    def __iter__(self) -> Iterator[BaseAssignment]:
        """Iterate through all assignments by ``for i in scope.assignments``."""
        for assignments in self._assignments.values():
            for assignment in assignments:
                yield assignment

    def __getitem__(self, node: Union[str, cst.CSTNode]) -> Collection[BaseAssignment]:
        """Get assignments given a name str or :class:`~libcst.CSTNode` by ``scope.assignments[node]``"""
        name = _NameUtil.get_name_for(node)
        return set(self._assignments[name]) if name in self._assignments else set()

    def __contains__(self, node: Union[str, cst.CSTNode]) -> bool:
        """Check if a name str or :class:`~libcst.CSTNode` has any assignment by ``node in scope.assignments``"""
        return len(self[node]) > 0


class Accesses:
    """A container to provide all accesses in a scope."""

    def __init__(self, accesses: Mapping[str, Collection[Access]]) -> None:
        self._accesses = accesses

    def __iter__(self) -> Iterator[Access]:
        """Iterate through all accesses by ``for i in scope.accesses``."""
        for accesses in self._accesses.values():
            for access in accesses:
                yield access

    def __getitem__(self, node: Union[str, cst.CSTNode]) -> Collection[Access]:
        """Get accesses given a name str or :class:`~libcst.CSTNode` by ``scope.accesses[node]``"""
        name = _NameUtil.get_name_for(node)
        return self._accesses[name] if name in self._accesses else set()

    def __contains__(self, node: Union[str, cst.CSTNode]) -> bool:
        """Check if a name str or :class:`~libcst.CSTNode` has any access by ``node in scope.accesses``"""
        return len(self[node]) > 0


class QualifiedNameSource(Enum):
    IMPORT = auto()
    BUILTIN = auto()
    LOCAL = auto()


@add_slots
@dataclass(frozen=True)
class QualifiedName:
    #: Qualified name, e.g. ``a.b.c`` or ``fn.<locals>.var``.
    name: str

    #: Source of the name, either :attr:`QualifiedNameSource.IMPORT`, :attr:`QualifiedNameSource.BUILTIN`
    #: or :attr:`QualifiedNameSource.LOCAL`.
    source: QualifiedNameSource


class _NameUtil:
    @staticmethod
    def get_name_for(node: Union[str, cst.CSTNode]) -> Optional[str]:
        """A helper function to retrieve simple name str from a CSTNode or str"""
        if isinstance(node, cst.Name):
            return node.value
        elif isinstance(node, str):
            return node
        elif isinstance(node, cst.Call):
            return _NameUtil.get_name_for(node.func)
        elif isinstance(node, cst.Subscript):
            return _NameUtil.get_name_for(node.value)
        elif isinstance(node, (cst.FunctionDef, cst.ClassDef)):
            return _NameUtil.get_name_for(node.name)
        return None

    @staticmethod
    def find_qualified_name_for_import_alike(
        assignment_node: Union[cst.Import, cst.ImportFrom], full_name: str
    ) -> Set[QualifiedName]:
        module = ""
        results = set()
        if isinstance(assignment_node, cst.ImportFrom):
            module_attr = assignment_node.module
            if module_attr:
                # TODO: for relative import, keep the relative Dot in the qualified name
                module = get_full_name_for_node(module_attr)
        import_names = assignment_node.names
        if not isinstance(import_names, cst.ImportStar):
            for name in import_names:
                real_name = get_full_name_for_node(name.name)
                if not real_name:
                    continue
                # real_name can contain `.` for dotted imports
                # for these we want to find the longest prefix that matches full_name
                parts = real_name.split(".")
                real_names = [".".join(parts[:i]) for i in range(len(parts), 0, -1)]
                for real_name in real_names:
                    as_name = real_name
                    if module:
                        real_name = f"{module}.{real_name}"
                    if name and name.asname:
                        eval_alias = name.evaluated_alias
                        if eval_alias is not None:
                            as_name = eval_alias
                    if full_name.startswith(as_name):
                        remaining_name = full_name.split(as_name, 1)[1].lstrip(".")
                        results.add(
                            QualifiedName(
                                f"{real_name}.{remaining_name}"
                                if remaining_name
                                else real_name,
                                QualifiedNameSource.IMPORT,
                            )
                        )
                        break
        return results

    @staticmethod
    def find_qualified_name_for_non_import(
        assignment: Assignment, remaining_name: str
    ) -> Set[QualifiedName]:
        scope = assignment.scope
        name_prefixes = []
        while scope:
            if isinstance(scope, ClassScope):
                name_prefixes.append(scope.name)
            elif isinstance(scope, FunctionScope):
                name_prefixes.append(f"{scope.name}.<locals>")
            elif isinstance(scope, GlobalScope):
                break
            elif isinstance(scope, ComprehensionScope):
                name_prefixes.append("<comprehension>")
            else:
                raise Exception(f"Unexpected Scope: {scope}")
            scope = scope.parent

        parts = [*reversed(name_prefixes)]
        if remaining_name:
            parts.append(remaining_name)
        return {QualifiedName(".".join(parts), QualifiedNameSource.LOCAL)}


class Scope(abc.ABC):
    """
    Base class of all scope classes. Scope object stores assignments from imports,
    variable assignments, function definition or class definition.
    A scope has a parent scope which represents the inheritance relationship. That means
    an assignment in parent scope is viewable to the child scope and the child scope may
    overwrites the assignment by using the same name.

    Use ``name in scope`` to check whether a name is viewable in the scope.
    Use ``scope[name]`` to retrieve all viewable assignments in the scope.

    .. note::
       This scope analysis module only analyzes local variable names and it doesn't handle
       attribute names; for example, given ``a.b.c = 1``, local variable name ``a`` is recorded
       as an assignment instead of ``c`` or ``a.b.c``. To analyze the assignment/access of
       arbitrary object attributes, we leave the job to type inference metadata provider
       coming in the future.
    """

    #: Parent scope. Note the parent scope of a GlobalScope is itself.
    parent: "Scope"

    #: Refers to the GlobalScope.
    globals: "GlobalScope"
    _assignments: MutableMapping[str, Set[BaseAssignment]]
    _accesses: MutableMapping[str, Set[Access]]
    _assignment_count: int

    def __init__(self, parent: "Scope") -> None:
        super().__init__()
        self.parent = parent
        self.globals = parent.globals
        self._assignments = defaultdict(set)
        self._accesses = defaultdict(set)
        self._assignment_count = 0

    def record_assignment(self, name: str, node: cst.CSTNode) -> None:
        self._assignments[name].add(
            Assignment(name=name, scope=self, node=node, index=self._assignment_count)
        )

    def record_access(self, name: str, access: Access) -> None:
        self._accesses[name].add(access)

    def _getitem_from_self_or_parent(self, name: str) -> Set[BaseAssignment]:
        """Overridden by ClassScope to hide it's assignments from child scopes."""
        return self[name]

    def _contains_in_self_or_parent(self, name: str) -> bool:
        """Overridden by ClassScope to hide it's assignments from child scopes."""
        return name in self

    def _record_assignment_as_parent(self, name: str, node: cst.CSTNode) -> None:
        """Overridden by ClassScope to forward 'nonlocal' assignments from child scopes."""
        self.record_assignment(name, node)

    @abc.abstractmethod
    def __contains__(self, name: str) -> bool:
        """ Check if the name str exist in current scope by ``name in scope``. """
        ...

    @abc.abstractmethod
    def __getitem__(self, name: str) -> Set[BaseAssignment]:
        """
        Get assignments given a name str by ``scope[name]``.

        .. note::
           *Why does it return a list of assignments given a name instead of just one assignment?*

           Many programming languages differentiate variable declaration and assignment.
           Further, those programming languages often disallow duplicate declarations within
           the same scope, and will often hoist the declaration (without its assignment) to
           the top of the scope. These design decisions make static analysis much easier,
           because it's possible to match a name against its single declaration for a given scope.

           As an example, the following code would be valid in JavaScript::

               function fn() {
                 console.log(value);  // value is defined here, because the declaration is hoisted, but is currently 'undefined'.
                 var value = 5;  // A function-scoped declaration.
               }
               fn();  // prints 'undefined'.

           In contrast, Python's declaration and assignment are identical and are not hoisted::

               if conditional_value:
                   value = 5
               elif other_conditional_value:
                   value = 10
               print(value)  # possibly valid, depending on conditional execution

           This code may throw a ``NameError`` if both conditional values are falsy.
           It also means that depending on the codepath taken, the original declaration
           could come from either ``value = ...`` assignment node.
           As a result, instead of returning a single declaration,
           we're forced to return a collection of all of the assignments we think could have
           defined a given name by the time a piece of code is executed.
           For the above example, value would resolve to a set of both assignments.
        """
        ...

    def __hash__(self) -> int:
        return id(self)

    @abc.abstractmethod
    def record_global_overwrite(self, name: str) -> None:
        ...

    @abc.abstractmethod
    def record_nonlocal_overwrite(self, name: str) -> None:
        ...

    def get_qualified_names_for(
        self, node: Union[str, cst.CSTNode]
    ) -> Collection[QualifiedName]:
        """Get all :class:`~libcst.metadata.QualifiedName` in current scope given a
        :class:`~libcst.CSTNode`.
        The source of a qualified name can be either :attr:`QualifiedNameSource.IMPORT`,
        :attr:`QualifiedNameSource.BUILTIN` or :attr:`QualifiedNameSource.LOCAL`.
        Given the following example, ``c`` has qualified name ``a.b.c`` with source ``IMPORT``,
        ``f`` has qualified name ``Cls.f`` with source ``LOCAL``, ``a`` has qualified name
        ``Cls.f.<locals>.a``, ``i`` has qualified name ``Cls.f.<locals>.<comprehension>.i``,
        and the builtin ``int`` has qualified name ``builtins.int`` with source ``BUILTIN``::

            from a.b import c
            class Cls:
                def f(self) -> "c":
                    c()
                    a = int("1")
                    [i for i in c()]

        We extends `PEP-3155 <https://www.python.org/dev/peps/pep-3155/>`_
        (defines ``__qualname__`` for class and function only; function namespace is followed
        by a ``<locals>``) to provide qualified name for all :class:`~libcst.CSTNode`
        recorded by :class:`~libcst.metadata.Assignment` and :class:`~libcst.metadata.Access`.
        The namespace of a comprehension (:class:`~libcst.ListComp`, :class:`~libcst.SetComp`,
        :class:`~libcst.DictComp`) is represented with ``<comprehension>``.

        An imported name may be used for type annotation with :class:`~libcst.SimpleString` and
        currently resolving the qualified given :class:`~libcst.SimpleString` is not supported
        considering it could be a complex type annotation in the string which is hard to
        resolve, e.g. ``List[Union[int, str]]``.
        """
        results = set()
        full_name = get_full_name_for_node(node)
        if full_name is None:
            return results
        assignments = set()
        parts = full_name.split(".")
        for i in range(len(parts), 0, -1):
            prefix = ".".join(parts[:i])
            if prefix in self:
                assignments = self[prefix]
                break
        for assignment in assignments:
            if isinstance(assignment, Assignment):
                assignment_node = assignment.node
                if isinstance(assignment_node, (cst.Import, cst.ImportFrom)):
                    results |= _NameUtil.find_qualified_name_for_import_alike(
                        assignment_node, full_name
                    )
                else:
                    results |= _NameUtil.find_qualified_name_for_non_import(
                        assignment, full_name
                    )
            elif isinstance(assignment, BuiltinAssignment):
                results.add(
                    QualifiedName(
                        f"builtins.{assignment.name}", QualifiedNameSource.BUILTIN
                    )
                )
        return results

    @property
    def assignments(self) -> Assignments:
        """Return an :class:`~libcst.metadata.Assignments` contains all assignmens in current scope."""
        return Assignments(self._assignments)

    @property
    def accesses(self) -> Accesses:
        """Return an :class:`~libcst.metadata.Accesses` contains all accesses in current scope."""
        return Accesses(self._accesses)


class GlobalScope(Scope):
    """
    A GlobalScope is the scope of module. All module level assignments are recorded in GlobalScope.
    """

    def __init__(self) -> None:
        self.globals: Scope = self  # must be defined before Scope.__init__ is called
        super().__init__(parent=self)

    def __contains__(self, name: str) -> bool:
        return hasattr(builtins, name) or (
            name in self._assignments and len(self._assignments[name]) > 0
        )

    def __getitem__(self, name: str) -> Set[BaseAssignment]:
        if hasattr(builtins, name):
            if not any(
                isinstance(i, BuiltinAssignment) for i in self._assignments[name]
            ):
                self._assignments[name].add(BuiltinAssignment(name, self))
        return self._assignments[name]

    def record_global_overwrite(self, name: str) -> None:
        pass

    def record_nonlocal_overwrite(self, name: str) -> None:
        raise NotImplementedError("nonlocal declaration not allowed at module level")


class LocalScope(Scope, abc.ABC):
    _scope_overwrites: Dict[str, Scope]

    #: Name of function. Used as qualified name.
    name: Optional[str]

    #: The :class:`~libcst.CSTNode` node defines the current scope.
    node: cst.CSTNode

    def __init__(
        self, parent: Scope, node: cst.CSTNode, name: Optional[str] = None
    ) -> None:
        super().__init__(parent)
        self.name = name
        self.node = node
        self._scope_overwrites = {}

    def record_global_overwrite(self, name: str) -> None:
        self._scope_overwrites[name] = self.globals

    def record_nonlocal_overwrite(self, name: str) -> None:
        self._scope_overwrites[name] = self.parent

    def record_assignment(self, name: str, node: cst.CSTNode) -> None:
        if name in self._scope_overwrites:
            self._scope_overwrites[name]._record_assignment_as_parent(name, node)
        else:
            super().record_assignment(name, node)

    def __contains__(self, name: str) -> bool:
        if name in self._scope_overwrites:
            return name in self._scope_overwrites[name]
        if name in self._assignments:
            return len(self._assignments[name]) > 0
        return self.parent._contains_in_self_or_parent(name)

    def __getitem__(self, name: str) -> Set[BaseAssignment]:
        if name in self._scope_overwrites:
            return self._scope_overwrites[name]._getitem_from_self_or_parent(name)
        if name in self._assignments:
            return self._assignments[name]
        else:
            return self.parent._getitem_from_self_or_parent(name)


# even though we don't override the constructor.
class FunctionScope(LocalScope):
    """
    When a function is defined, it creates a FunctionScope.
    """

    pass


# even though we don't override the constructor.
class ClassScope(LocalScope):
    """
    When a class is defined, it creates a ClassScope.
    """

    def _record_assignment_as_parent(self, name: str, node: cst.CSTNode) -> None:
        """
        Forward the assignment to parent.

            def outer_fn():
                v = ...  # outer_fn's declaration
                class InnerCls:
                    v = ...  # shadows outer_fn's declaration
                    def inner_fn():
                        nonlocal v
                        v = ...  # this should actually refer to outer_fn's declaration
                                 # and not to InnerCls's, because InnerCls's scope is
                                 # hidden from its children.

        """
        self.parent._record_assignment_as_parent(name, node)

    def _getitem_from_self_or_parent(self, name: str) -> Set[BaseAssignment]:
        """
        Class variables are only accessible using ClassName.attribute, cls.attribute, or
        self.attribute in child scopes. They cannot be accessed with their bare names.
        """
        return self.parent._getitem_from_self_or_parent(name)

    def _contains_in_self_or_parent(self, name: str) -> bool:
        """
        See :meth:`_getitem_from_self_or_parent`
        """
        return self.parent._contains_in_self_or_parent(name)


# even though we don't override the constructor.
class ComprehensionScope(LocalScope):
    """
    Comprehensions and generator expressions create their own scope. For example, in

        [i for i in range(10)]

    The variable ``i`` is only viewable within the ComprehensionScope.
    """

    # TODO: Assignment expressions (Python 3.8) will complicate ComprehensionScopes,
    # and will require us to handle such assignments as non-local.
    # https://www.python.org/dev/peps/pep-0572/#scope-of-the-target
    pass


# Generates dotted names from an Attribute or Name node:
# Attribute(value=Name(value="a"), attr=Name(value="b")) -> ("a.b", "a")
# each string has the corresponding CSTNode attached to it
def _gen_dotted_names(
    node: Union[cst.Attribute, cst.Name]
) -> Iterator[Tuple[str, Union[cst.Attribute, cst.Name]]]:
    if isinstance(node, cst.Name):
        yield node.value, node
    else:
        value = node.value
        if isinstance(value, cst.Call):
            value = value.func
            if isinstance(value, (cst.Attribute, cst.Name)):
                name_values = _gen_dotted_names(value)
                try:
                    next_name, next_node = next(name_values)
                except StopIteration:
                    return
                else:
                    yield next_name, next_node
                    yield from name_values
        elif isinstance(value, (cst.Attribute, cst.Name)):
            name_values = _gen_dotted_names(value)
            try:
                next_name, next_node = next(name_values)
            except StopIteration:
                return
            else:
                yield f"{next_name}.{node.attr.value}", node
                yield next_name, next_node
                yield from name_values


class ScopeVisitor(cst.CSTVisitor):
    # since it's probably not useful. That can makes this visitor cleaner.
    def __init__(self, provider: "ScopeProvider") -> None:
        self.provider: ScopeProvider = provider
        self.scope: Scope = GlobalScope()
        self.__deferred_accesses: List[Tuple[Access, Optional[cst.Attribute]]] = []
        self.__top_level_attribute_stack: List[Optional[cst.Attribute]] = [None]
        self.__in_annotation: Set[
            Union[cst.Call, cst.Annotation, cst.Subscript]
        ] = set()
        self.__in_type_hint: Set[Union[cst.Call, cst.Annotation, cst.Subscript]] = set()
        self.__in_ignored_subscript: Set[cst.Subscript] = set()
        self.__ignore_annotation: int = 0

    @contextmanager
    def _new_scope(
        self, kind: Type[LocalScope], node: cst.CSTNode, name: Optional[str] = None
    ) -> Iterator[None]:
        parent_scope = self.scope
        self.scope = kind(parent_scope, node, name)
        try:
            yield
        finally:
            self.scope = parent_scope

    @contextmanager
    def _switch_scope(self, scope: Scope) -> Iterator[None]:
        current_scope = self.scope
        self.scope = scope
        try:
            yield
        finally:
            self.scope = current_scope

    def _visit_import_alike(self, node: Union[cst.Import, cst.ImportFrom]) -> bool:
        names = node.names
        if isinstance(names, cst.ImportStar):
            return False

        # make sure node.names is Sequence[ImportAlias]
        for name in names:
            self.provider.set_metadata(name, self.scope)
            asname = name.asname
            if asname is not None:
                name_values = _gen_dotted_names(cst.ensure_type(asname.name, cst.Name))
            else:
                name_values = _gen_dotted_names(name.name)

            for name_value, _ in name_values:
                self.scope.record_assignment(name_value, node)
        return False

    def visit_Import(self, node: cst.Import) -> Optional[bool]:
        return self._visit_import_alike(node)

    def visit_ImportFrom(self, node: cst.ImportFrom) -> Optional[bool]:
        return self._visit_import_alike(node)

    def visit_Attribute(self, node: cst.Attribute) -> Optional[bool]:
        if self.__top_level_attribute_stack[-1] is None:
            self.__top_level_attribute_stack[-1] = node
        node.value.visit(self)  # explicitly not visiting attr
        if self.__top_level_attribute_stack[-1] is node:
            self.__top_level_attribute_stack[-1] = None
        return False

    def visit_Call(self, node: cst.Call) -> Optional[bool]:
        self.__top_level_attribute_stack.append(None)
        qnames = {qn.name for qn in self.scope.get_qualified_names_for(node)}
        if "typing.NewType" in qnames or "typing.TypeVar" in qnames:
            node.func.visit(self)
            self.__in_type_hint.add(node)
            for arg in node.args[1:]:
                arg.visit(self)
            return False
        if "typing.cast" in qnames:
            node.func.visit(self)
            if len(node.args) > 0:
                self.__in_type_hint.add(node)
                node.args[0].visit(self)
                self.__in_type_hint.discard(node)
                for arg in node.args[1:]:
                    arg.visit(self)
            return False
        return True

    def leave_Call(self, original_node: cst.Call) -> None:
        self.__top_level_attribute_stack.pop()
        self.__in_type_hint.discard(original_node)

    def visit_Annotation(self, node: cst.Annotation) -> Optional[bool]:
        self.__in_annotation.add(node)

    def leave_Annotation(self, original_node: cst.Annotation) -> None:
        self.__in_annotation.discard(original_node)

    def visit_SimpleString(self, node: cst.SimpleString) -> Optional[bool]:
        self._handle_string_annotation(node)
        return False

    def visit_ConcatenatedString(self, node: cst.ConcatenatedString) -> Optional[bool]:
        return not self._handle_string_annotation(node)

    def _handle_string_annotation(
        self, node: Union[cst.SimpleString, cst.ConcatenatedString]
    ) -> bool:
        """Returns whether it successfully handled the string annotation"""
        if (
            self.__in_type_hint or self.__in_annotation
        ) and not self.__in_ignored_subscript:
            value = node.evaluated_value
            if value:
                mod = cst.parse_module(value)
                mod.visit(self)
                return True
        return False

    def visit_Subscript(self, node: cst.Subscript) -> Optional[bool]:
        if isinstance(node.value, cst.Name):
            qnames = {qn.name for qn in self.scope.get_qualified_names_for(node.value)}
            if any(qn.startswith(("typing.", "typing_extensions.")) for qn in qnames):
                self.__in_type_hint.add(node)
            if "typing.Literal" in qnames or "typing_extensions.Literal" in qnames:
                self.__in_ignored_subscript.add(node)
        return True

    def leave_Subscript(self, original_node: cst.Subscript) -> None:
        self.__in_type_hint.discard(original_node)
        self.__in_ignored_subscript.discard(original_node)

    def visit_Name(self, node: cst.Name) -> Optional[bool]:
        # not all Name have ExpressionContext
        context = self.provider.get_metadata(ExpressionContextProvider, node, None)
        if context == ExpressionContext.STORE:
            self.scope.record_assignment(node.value, node)
        elif context in (ExpressionContext.LOAD, ExpressionContext.DEL, None):
            access = Access(
                node,
                self.scope,
                is_annotation=bool(
                    self.__in_annotation and not self.__ignore_annotation
                ),
                is_type_hint=bool(self.__in_type_hint),
            )
            self.__deferred_accesses.append(
                (access, self.__top_level_attribute_stack[-1])
            )

    def visit_FunctionDef(self, node: cst.FunctionDef) -> Optional[bool]:
        self.scope.record_assignment(node.name.value, node)
        self.provider.set_metadata(node.name, self.scope)

        with self._new_scope(FunctionScope, node, get_full_name_for_node(node.name)):
            node.params.visit(self)
            node.body.visit(self)

        for decorator in node.decorators:
            decorator.visit(self)
        returns = node.returns
        if returns:
            returns.visit(self)

        return False

    def visit_Lambda(self, node: cst.Lambda) -> Optional[bool]:
        with self._new_scope(FunctionScope, node):
            node.params.visit(self)
            node.body.visit(self)
        return False

    def visit_Param(self, node: cst.Param) -> Optional[bool]:
        self.scope.record_assignment(node.name.value, node)
        self.provider.set_metadata(node.name, self.scope)
        with self._switch_scope(self.scope.parent):
            for field in [node.default, node.annotation]:
                if field:
                    field.visit(self)

        return False

    def visit_Arg(self, node: cst.Arg) -> bool:
        # The keyword of Arg is neither an Assignment nor an Access and we explicitly don't visit it.
        value = node.value
        if value:
            value.visit(self)
        return False

    def visit_ClassDef(self, node: cst.ClassDef) -> Optional[bool]:
        self.scope.record_assignment(node.name.value, node)
        self.provider.set_metadata(node.name, self.scope)
        for decorator in node.decorators:
            decorator.visit(self)
        for base in node.bases:
            base.visit(self)
        for keyword in node.keywords:
            keyword.visit(self)

        with self._new_scope(ClassScope, node, get_full_name_for_node(node.name)):
            for statement in node.body.body:
                statement.visit(self)
        return False

    def visit_ClassDef_bases(self, node: cst.ClassDef) -> None:
        self.__ignore_annotation += 1

    def leave_ClassDef_bases(self, node: cst.ClassDef) -> None:
        self.__ignore_annotation -= 1

    def visit_Global(self, node: cst.Global) -> Optional[bool]:
        for name_item in node.names:
            self.scope.record_global_overwrite(name_item.name.value)
        return False

    def visit_Nonlocal(self, node: cst.Nonlocal) -> Optional[bool]:
        for name_item in node.names:
            self.scope.record_nonlocal_overwrite(name_item.name.value)
        return False

    def visit_ListComp(self, node: cst.ListComp) -> Optional[bool]:
        return self._visit_comp_alike(node)

    def visit_SetComp(self, node: cst.SetComp) -> Optional[bool]:
        return self._visit_comp_alike(node)

    def visit_DictComp(self, node: cst.DictComp) -> Optional[bool]:
        return self._visit_comp_alike(node)

    def visit_GeneratorExp(self, node: cst.GeneratorExp) -> Optional[bool]:
        return self._visit_comp_alike(node)

    def _visit_comp_alike(
        self, node: Union[cst.ListComp, cst.SetComp, cst.DictComp, cst.GeneratorExp]
    ) -> bool:
        """
        Cheat sheet: `[elt for target in iter if ifs]`

        Terminology:
            target: The variable or pattern we're storing each element of the iter in.
            iter: The thing we're iterating over.
            ifs: A list of conditions provided
            elt: The value that will be computed and "yielded" each time the loop
                iterates. For most comprehensions, this is just the `node.elt`, but
                DictComp has `key` and `value`, which behave like `node.elt` would.


        Nested Comprehension: ``[a for b in c for a in b]`` is a "nested" ListComp.
        The outer iterator is in ``node.for_in`` and the inner iterator is in
        ``node.for_in.inner_for_in``.


        The first comprehension object's iter in generators is evaluated
        outside of the ComprehensionScope. Every other comprehension's iter is
        evaluated inside the ComprehensionScope. Even though that doesn't seem very sane,
        but that appears to be how it works.

            non_flat = [ [1,2,3], [4,5,6], [7,8]
            flat = [y for x in non_flat for y in x]  # this works fine

            # This will give a "NameError: name 'x' is not defined":
            flat = [y for x in x for y in x]
            # x isn't defined, because the first iter is evaluted outside the scope.

            # This will give an UnboundLocalError, indicating that the second
            # comprehension's iter value is evaluated inside the scope as its elt.
            # UnboundLocalError: local variable 'y' referenced before assignment
            flat = [y for x in non_flat for y in y]
        """
        for_in = node.for_in
        for_in.iter.visit(self)
        self.provider.set_metadata(for_in, self.scope)
        with self._new_scope(ComprehensionScope, node):
            for_in.target.visit(self)
            # Things from here on can refer to the target.
            self.scope._assignment_count += 1
            for condition in for_in.ifs:
                condition.visit(self)
            inner_for_in = for_in.inner_for_in
            if inner_for_in:
                inner_for_in.visit(self)
            if isinstance(node, cst.DictComp):
                node.key.visit(self)
                node.value.visit(self)
            else:
                node.elt.visit(self)
        return False

    def visit_For(self, node: cst.For) -> Optional[bool]:
        node.target.visit(self)
        self.scope._assignment_count += 1
        for child in [node.iter, node.body, node.orelse, node.asynchronous]:
            if child is not None:
                child.visit(self)
        return False

    def infer_accesses(self) -> None:
        # Aggregate access with the same name and batch add with set union as an optimization.
        # In worst case, all accesses (m) and assignments (n) refer to the same name,
        # the time complexity is O(m x n), this optimizes it as O(m + n).
        scope_name_accesses = defaultdict(set)
        for (access, enclosing_attribute) in self.__deferred_accesses:
            name = ensure_type(access.node, cst.Name).value
            if enclosing_attribute is not None:
                # if _gen_dotted_names doesn't generate any values, fall back to
                # the original name node above
                for attr_name, node in _gen_dotted_names(enclosing_attribute):
                    if attr_name in access.scope:
                        access.node = node
                        name = attr_name
                        break

            scope_name_accesses[(access.scope, name)].add(access)
            access.record_assignments(name)
            access.scope.record_access(name, access)

        for (scope, name), accesses in scope_name_accesses.items():
            for assignment in scope[name]:
                assignment.record_accesses(accesses)

        self.__deferred_accesses = []

    def on_leave(self, original_node: cst.CSTNode) -> None:
        self.provider.set_metadata(original_node, self.scope)
        if isinstance(original_node, _ASSIGNMENT_LIKE_NODES):
            self.scope._assignment_count += 1
        super().on_leave(original_node)


class ScopeProvider(BatchableMetadataProvider[Optional[Scope]]):
    """
    :class:`ScopeProvider` traverses the entire module and creates the scope inheritance
    structure. It provides the scope of name assignment and accesses. It is useful for
    more advanced static analysis. E.g. given a :class:`~libcst.FunctionDef`
    node, we can check the type of its Scope to figure out whether it is a class method
    (:class:`ClassScope`) or a regular function (:class:`GlobalScope`).

    Scope metadata is available for most node types other than formatting information nodes
    (whitespace, parentheses, etc.).
    """

    METADATA_DEPENDENCIES = (ExpressionContextProvider,)

    def visit_Module(self, node: cst.Module) -> Optional[bool]:
        visitor = ScopeVisitor(self)
        node.visit(visitor)
        visitor.infer_accesses()


class QualifiedNameVisitor(cst.CSTVisitor):
    def __init__(self, provider: "QualifiedNameProvider") -> None:
        self.provider: QualifiedNameProvider = provider

    def on_visit(self, node: cst.CSTNode) -> bool:
        scope = self.provider.get_metadata(ScopeProvider, node, None)
        if scope:
            self.provider.set_metadata(node, scope.get_qualified_names_for(node))
        else:
            self.provider.set_metadata(node, set())
        super().on_visit(node)
        return True


class QualifiedNameProvider(BatchableMetadataProvider[Collection[QualifiedName]]):
    """
    Compute possible qualified names of a variable CSTNode
    (extends `PEP-3155 <https://www.python.org/dev/peps/pep-3155/>`_).
    It uses the
    :func:`~libcst.metadata.Scope.get_qualified_names_for` underlying to get qualified names.
    Multiple qualified names may be returned, such as when we have conditional imports or an
    import shadows another. E.g., the provider finds ``a.b``, ``d.e`` and
    ``f.g`` as possible qualified names of ``c``::

        >>> wrapper = MetadataWrapper(
        >>>     cst.parse_module(dedent(
        >>>     '''
        >>>         if something:
        >>>             from a import b as c
        >>>         elif otherthing:
        >>>             from d import e as c
        >>>         else:
        >>>             from f import g as c
        >>>         c()
        >>>     '''
        >>>     ))
        >>> )
        >>> call = wrapper.module.body[1].body[0].value
        >>> wrapper.resolve(QualifiedNameProvider)[call],
        {
            QualifiedName(name="a.b", source=QualifiedNameSource.IMPORT),
            QualifiedName(name="d.e", source=QualifiedNameSource.IMPORT),
            QualifiedName(name="f.g", source=QualifiedNameSource.IMPORT),
        }

    For qualified name of a variable in a function or a comprehension, please refer
    :func:`~libcst.metadata.Scope.get_qualified_names_for` for more detail.
    """

    METADATA_DEPENDENCIES = (ScopeProvider,)

    def visit_Module(self, node: cst.Module) -> Optional[bool]:
        visitor = QualifiedNameVisitor(self)
        node.visit(visitor)

    @staticmethod
    def has_name(
        visitor: MetadataDependent, node: cst.CSTNode, name: Union[str, QualifiedName]
    ) -> bool:
        """Check if any of qualified name has the str name or :class:`~libcst.metadata.QualifiedName` name."""
        qualified_names = visitor.get_metadata(QualifiedNameProvider, node, set())
        if isinstance(name, str):
            return any(qn.name == name for qn in qualified_names)
        else:
            return any(qn == name for qn in qualified_names)



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#
from libcst import Module
from libcst.codemod import CodemodCommand


class NOOPCommand(CodemodCommand):
    DESCRIPTION: str = "Does absolutely nothing."

    def transform_module_impl(self, tree: Module) -> Module:
        # Return the tree as-is, with absolutely no modification
        return tree



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Callable, Optional, Sequence, TypeVar

from libcst._parser.types.conversions import NonterminalConversion
from libcst._parser.types.production import Production


_NonterminalConversionT = TypeVar(
    "_NonterminalConversionT", bound=NonterminalConversion
)


# We could version our grammar at a later point by adding a version metadata kwarg to
# this decorator.
def with_production(
    production_name: str,
    children: str,
    *,
    version: Optional[str] = None,
    future: Optional[str] = None,
) -> Callable[[_NonterminalConversionT], _NonterminalConversionT]:
    """
    Attaches a bit of grammar to a conversion function. The parser extracts all of these
    production strings, and uses it to form the language's full grammar.

    If you need to attach multiple productions to the same conversion function
    """

    def inner(fn: _NonterminalConversionT) -> _NonterminalConversionT:
        if not hasattr(fn, "productions"):
            fn.productions = []
        # pyre-ignore: Pyre doesn't think that fn has a __name__ attribute
        fn_name = fn.__name__
        if not fn_name.startswith("convert_"):
            raise Exception(
                "A function with a production must be named 'convert_X', not "
                + f"'{fn_name}'."
            )
        # pyre-ignore: Pyre doesn't know about this magic field we added
        fn.productions.append(Production(production_name, children, version, future))
        return fn

    return inner


def get_productions(fn: NonterminalConversion) -> Sequence[Production]:
    # pyre-ignore Pyre doesn't know about this magic field we added
    return fn.productions



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
# pyre-unsafe

import inspect
import re
from functools import wraps
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
)
from unittest import TestCase


DATA_PROVIDER_DATA_ATTR_NAME = "__data_provider_data"
DATA_PROVIDER_DESCRIPTION_PREFIX = "_data_provider_"
PROVIDER_TEST_LIMIT_ATTR_NAME = "__provider_test_limit"
DEFAULT_TEST_LIMIT = 256


T = TypeVar("T")


def none_throws(value: Optional[T], message: str = "Unexpected None value") -> T:
    assert value is not None, message
    return value


def update_test_limit(test_method: Any, test_limit: int) -> None:
    # Store the maximum number of generated tests on the test_method. Since
    # contextmanager_provider can be specified multiple times, we need to
    # take the maximum of the existing attribute and the current value
    existing_test_limit = getattr(
        test_method, PROVIDER_TEST_LIMIT_ATTR_NAME, test_limit
    )
    setattr(
        test_method, PROVIDER_TEST_LIMIT_ATTR_NAME, max(existing_test_limit, test_limit)
    )


def try_get_provider_attr(
    member_name: str, member: Any, attr_name: str
) -> Optional[Any]:
    if inspect.isfunction(member) and member_name.startswith("test"):
        return getattr(member, attr_name, None)
    return None


def populate_data_provider_tests(dct: Dict[str, Any]) -> None:
    test_methods_to_add: Dict[str, Callable] = {}
    test_methods_to_remove: List[str] = []
    for member_name, member in dct.items():
        provider_data = try_get_provider_attr(
            member_name, member, DATA_PROVIDER_DATA_ATTR_NAME
        )
        if provider_data is not None:

            for description, data in (
                provider_data.items()
                if isinstance(provider_data, dict)
                else enumerate(provider_data)
            ):
                if isinstance(provider_data, dict):
                    description = f"{DATA_PROVIDER_DESCRIPTION_PREFIX}{description}"

                assert re.fullmatch(
                    r"[a-zA-Z0-9_]+", str(description)
                ), f"Testcase description must be a valid python identifier: '{description}'"

                @wraps(member)
                def new_test(
                    self: object,
                    data: Iterable[object] = data,
                    member: Callable[..., object] = member,
                ) -> object:
                    if isinstance(data, dict):
                        return member(self, **data)
                    else:
                        return member(self, *data)

                name = f"{member_name}_{description}"
                new_test.__name__ = name
                test_methods_to_add[name] = new_test
            if not test_methods_to_add:
                raise ValueError(
                    f"No data_provider tests were created for {member_name}! Please double check your data."
                )
            test_methods_to_remove.append(member_name)
    dct.update(test_methods_to_add)

    # Remove all old methods
    for test_name in test_methods_to_remove:
        del dct[test_name]


def validate_provider_tests(dct: Dict[str, Any]) -> None:
    members_to_replace = {}

    for member_name, member in dct.items():
        test_limit = try_get_provider_attr(
            member_name, member, PROVIDER_TEST_LIMIT_ATTR_NAME
        )
        if test_limit is not None:
            data = try_get_provider_attr(
                member_name, member, DATA_PROVIDER_DATA_ATTR_NAME
            )
            num_tests = len(data) if data else 1

            if num_tests > test_limit:
                # We don't use wraps() here so that the test isn't expanded
                # as it normally would be by whichever provider it uses
                def test_replacement(
                    self: Any,
                    member_name: Any = member_name,
                    num_tests: Any = num_tests,
                    test_limit: Any = test_limit,
                ) -> None:
                    raise AssertionError(
                        f"{member_name} generated {num_tests} tests but the limit is "
                        + f"{test_limit}. You can increase the number of "
                        + "allowed tests by specifying test_limit, but please "
                        + "consider whether you really need to test all of "
                        + "these combinations."
                    )

                test_replacement.__name__ = member_name
                members_to_replace[member_name] = test_replacement

    for member_name, new_member in members_to_replace.items():
        dct[member_name] = new_member


TestCaseType = Union[Sequence[object], Mapping[str, object]]
# Can't use Sequence[TestCaseType] here as some clients may pass in a Generator[TestCaseType]
StaticDataType = Union[Iterable[TestCaseType], Mapping[str, TestCaseType]]


def data_provider(
    static_data: StaticDataType, *, test_limit: int = DEFAULT_TEST_LIMIT
) -> Callable[[Callable], Callable]:
    # We need to be able to iterate over static_data more than once
    # (for validation), so if we weren't passed in a dict, list, or tuple
    # then we'll just create a list from the data
    if not isinstance(static_data, (dict, list, tuple)):
        static_data = list(static_data)

    def test_decorator(test_method: Callable) -> Callable:
        update_test_limit(test_method, test_limit)

        setattr(test_method, DATA_PROVIDER_DATA_ATTR_NAME, static_data)
        return test_method

    return test_decorator


class BaseTestMeta(type):
    def __new__(mcs, name: str, bases: Tuple[type, ...], dct: Dict[str, Any]) -> object:
        validate_provider_tests(dct)
        populate_data_provider_tests(dct)
        return super().__new__(mcs, name, bases, dict(dct))


class UnitTest(TestCase, metaclass=BaseTestMeta):
    pass



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Callable, Optional

import libcst as cst
from libcst import parse_expression
from libcst._nodes.tests.base import CSTNodeTest
from libcst.metadata import CodeRange
from libcst.testing.utils import data_provider


class IfExpTest(CSTNodeTest):
    @data_provider(
        (
            # Simple if experessions
            (
                cst.IfExp(
                    body=cst.Name("foo"), test=cst.Name("bar"), orelse=cst.Name("baz")
                ),
                "foo if bar else baz",
            ),
            # Parenthesized if expressions
            (
                cst.IfExp(
                    lpar=(cst.LeftParen(),),
                    body=cst.Name("foo"),
                    test=cst.Name("bar"),
                    orelse=cst.Name("baz"),
                    rpar=(cst.RightParen(),),
                ),
                "(foo if bar else baz)",
            ),
            (
                cst.IfExp(
                    body=cst.Name(
                        "foo", lpar=(cst.LeftParen(),), rpar=(cst.RightParen(),)
                    ),
                    whitespace_before_if=cst.SimpleWhitespace(""),
                    whitespace_after_if=cst.SimpleWhitespace(""),
                    test=cst.Name(
                        "bar", lpar=(cst.LeftParen(),), rpar=(cst.RightParen(),)
                    ),
                    whitespace_before_else=cst.SimpleWhitespace(""),
                    whitespace_after_else=cst.SimpleWhitespace(""),
                    orelse=cst.Name(
                        "baz", lpar=(cst.LeftParen(),), rpar=(cst.RightParen(),)
                    ),
                ),
                "(foo)if(bar)else(baz)",
                CodeRange((1, 0), (1, 21)),
            ),
            # Make sure that spacing works
            (
                cst.IfExp(
                    lpar=(cst.LeftParen(whitespace_after=cst.SimpleWhitespace(" ")),),
                    body=cst.Name("foo"),
                    whitespace_before_if=cst.SimpleWhitespace("  "),
                    whitespace_after_if=cst.SimpleWhitespace("  "),
                    test=cst.Name("bar"),
                    whitespace_before_else=cst.SimpleWhitespace("  "),
                    whitespace_after_else=cst.SimpleWhitespace("  "),
                    orelse=cst.Name("baz"),
                    rpar=(cst.RightParen(whitespace_before=cst.SimpleWhitespace(" ")),),
                ),
                "( foo  if  bar  else  baz )",
                CodeRange((1, 2), (1, 25)),
            ),
        )
    )
    def test_valid(
        self, node: cst.CSTNode, code: str, position: Optional[CodeRange] = None
    ) -> None:
        self.validate_node(node, code, parse_expression, expected_position=position)

    @data_provider(
        (
            (
                lambda: cst.IfExp(
                    cst.Name("bar"),
                    cst.Name("foo"),
                    cst.Name("baz"),
                    lpar=(cst.LeftParen(),),
                ),
                "left paren without right paren",
            ),
            (
                lambda: cst.IfExp(
                    cst.Name("bar"),
                    cst.Name("foo"),
                    cst.Name("baz"),
                    rpar=(cst.RightParen(),),
                ),
                "right paren without left paren",
            ),
        )
    )
    def test_invalid(
        self, get_node: Callable[[], cst.CSTNode], expected_re: str
    ) -> None:
        self.assert_invalid(get_node, expected_re)
